"""
Linear models. This module is in very early development and is subject to frequent breaking changes.

This module requires the NumPy package. PDS only requires Polars, but you can get all the optional dependencies by

`pip install polars_ds[all]`

"""

# Currently skipping tests for this module because the underlying functions are all tested in one way or
# another.

from __future__ import annotations

import polars as pl
import numpy as np
from typing import List, Tuple, Literal
from .typing import LRSolverMethods, NullPolicy, PolarsFrame, TypeAlias

from polars_ds._polars_ds import PyLR, PyGLM, PyElasticNet, PyOnlineLR

import sys

if sys.version_info >= (3, 11):
    from typing import Self
else:  # 3.10, 3.9, 3.8
    from typing_extensions import Self


# --------------------------------------------------------------------------------


def _handle_nulls_in_df(
    df: PolarsFrame, features: List[str], target: str, null_policy: NullPolicy
) -> PolarsFrame:
    if null_policy == "ignore":
        return df
    elif null_policy == "raise":
        total_null_count = (
            df.lazy().select(*features, target).null_count().collect().sum_horizontal()[0]
        )
        if total_null_count > 0:
            raise ValueError("Nulls found in Dataframe.")
        return df
    elif null_policy == "skip":
        return df.drop_nulls(subset=features + [target])
    elif null_policy == "zero":
        return df.with_columns(pl.col(features).fill_null(0.0)).drop_nulls(subset=target)
    elif null_policy == "one":
        return df.with_columns(pl.col(features).fill_null(1.0)).drop_nulls(subset=target)
    else:
        try:
            fill_value = float(null_policy)
            if np.isfinite(fill_value):
                return df.with_columns(pl.col(features).fill_null(fill_value)).drop_nulls(
                    subset=target
                )
            raise ValueError("When null_policy is a number, it cannot be nan or infinite.")
        except Exception as e:
            raise ValueError(f"Unknown null_policy. Error: {e}")


def _handle_nans_in_np(
    X: np.ndarray,  # N x M
    y: np.ndarray,  # N x 1
    null_policy: NullPolicy,
) -> Tuple[np.ndarray, np.ndarray]:
    if null_policy == "ignore":
        return (X, y)
    elif null_policy == "raise":
        if np.any(np.isnan(X)) | np.any(np.isnan(y)):
            raise ValueError("Nulls found in X or y.")
        return (X, y)
    elif null_policy == "skip":
        row_has_nan = np.any(np.isnan(X), axis=1) | np.any(np.isnan(y), axis=1)
        return (X[~row_has_nan], y[~row_has_nan])
    elif null_policy == "zero":
        y_nans = np.any(np.isnan(y), axis=1)
        return (np.nan_to_num(X, nan=0.0)[~y_nans], y[~y_nans])
    elif null_policy == "one":
        y_nans = np.any(np.isnan(y), axis=1)
        return (np.nan_to_num(X, nan=1.0)[~y_nans], y[~y_nans])
    else:
        try:
            fill_value = float(null_policy)
            if np.isfinite(fill_value):
                y_nans = np.any(np.isnan(y), axis=1)
                return (np.nan_to_num(X, nan=fill_value)[~y_nans], y[~y_nans])
            raise ValueError("When null_policy is a number, it cannot be nan or infinite.")
        except Exception as e:
            raise ValueError(f"Unknown null_policy. Error: {e}")


# --------------------------------------------------------------------------------


class LR:
    """
    Normal or Ridge Regression.
    """

    def __init__(
        self,
        has_bias: bool = False,
        lambda_: float = 0.0,
        solver: LRSolverMethods = "qr",
        feature_names_in_: List[str] | None = None,
    ):
        """
        Parameters
        ----------
        lambda_
            The regularization parameters for ridge. If this is positive, then this class will solve Ridge.
        solver
            Use one of 'svd', 'cholesky' and 'qr' method to solve the least square equation. Default is 'qr'.
        has_bias
            Whether to add a bias term. Also known as intercept in other packages.
        feature_names_in_
            Names for the incoming features, if available. If None, the names will be empty. They will be
            learned if .fit_df() is run later, or .set_input_features() is set later.
        """
        self._lr = PyLR(solver, lambda_, has_bias)
        self.feature_names_in_: List[str] = (
            [] if feature_names_in_ is None else list(feature_names_in_)
        )

    @classmethod
    def from_values(
        cls, coeffs: List[float], bias: float = 0.0, feature_names_in_: List[str] | None = None
    ) -> Self:
        """
        Constructs a LR class instance from coefficients and bias values.

        Parameters
        ----------
        coeffs
            Iterable of numbers representing the coefficients
        bias
            Value for the bias term
        feature_names_in_
            Names for the incoming features, if available. If None, the names will be empty. They will be
            learned if .fit_df() is run later, or .set_input_features() is set later.
        """
        coefficients = np.ascontiguousarray(coeffs, dtype=np.float64).flatten()
        lr = cls(
            has_bias=(bias != 0.0),
            lambda_=0.0,
            solver="Not Solved",
            feature_names_in_=feature_names_in_,
        )
        lr._lr.set_coeffs_and_bias(coefficients, bias)
        return lr

    def is_fit(self) -> bool:
        return self._lr.is_fit()

    def __repr__(self) -> str:
        if self._lr.lambda_ > 0.0:
            output = "Linear Regression (Ridge) Model\n"
        else:
            output = "Linear Regression Model\n"

        if self._lr.is_fit():
            output += f"Coefficients: {list(round(x, 5) for x in self._lr.coeffs)}\n"
            output += f"Bias/Intercept: {self._lr.bias}\n"
        else:
            output += "Not fitted yet."
        return output

    def set_input_features(self, features: List[str]) -> Self:
        """
        Sets the names of input features.

        Parameters
        ----------
        features
            List of strings.
        """
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self

    def coeffs(self) -> np.ndarray:
        """
        Returns a copy of the coefficients.
        """
        return self._lr.coeffs

    def bias(self) -> float:
        self._lr.bias

    def fit(self, X: np.ndarray, y: np.ndarray, null_policy: NullPolicy = "ignore") -> Self:
        """
        Fit the linear regression model on NumPy data.

        Parameters
        ----------
        X
            The feature Matrix. NumPy 2D matrix only.
        y
            The target data. NumPy array. Must be reshape-able to (-1, 1).
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.
        """
        self._lr.fit(*_handle_nans_in_np(X, y.reshape((-1, 1)), null_policy))
        return self

    def fit_df(
        self,
        df: PolarsFrame,
        features: List[str],
        target: str,
        null_policy: NullPolicy = "skip",
        show_report: bool = False,
    ) -> Self:
        """
        Fit the linear regression model on a dataframe. This will overwrite previously set feature names.
        The null policy only handles null values in df, not NaN values. It is the user's responsibility to handle
        NaN values if they exist in their pipeline.

        Parameters
        ----------
        df
            Either an eager or a lazy Polars dataframe.
        features
            List of strings of column names.
        target
            The target column's name.
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.
        show_report
            Whether to print out a regression report. This will duplicate work and will not work for Ridge
            regression. E.g. Nothing will be printed if lambda_ > 0.
        """
        if show_report and self._lr.lambda_ == 0.0:
            from . import query_lstsq_report

            print(
                df.lazy()
                .select(
                    query_lstsq_report(
                        *features,
                        target=target,
                    ).alias("report")
                )
                .unnest("report")
                .collect()
            )

        df2 = (
            _handle_nulls_in_df(df.lazy(), features, target, null_policy)
            .select(*features, target)
            .collect()
        )
        X = df2.select(features).to_numpy()
        y = df2.select(target).to_numpy()
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self.fit(X, y)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Returns the prediction of this linear model.

        Parameters
        ----------
        X
            Data to predict on, as a matrix
        """
        return self._lr.predict(X).reshape((-1, 1))

    def predict_df(self, df: PolarsFrame, name: str = "prediction") -> PolarsFrame:
        """
        Computes the prediction of the linear model and append it as a column in the dataframe. If input
        is lazy, output will be lazy.

        Parameters
        ----------
        df
            Either an eager or a lazy Polars dataframe.
        name
            The name of the prediction column
        """
        if len(self.feature_names_in_) <= 0:
            raise ValueError(
                "The linear model is not fitted on a dataframe, or no feature names have been given."
                "Not enough info to predict on a dataframe. Hint: try .fit_df() or .set_input_features()."
            )

        pred = pl.sum_horizontal(
            beta * pl.col(c) for c, beta in zip(self.feature_names_in_, self._lr.coeffs)
        )
        bias = self._lr.bias
        if bias != 0.0:
            pred = pred + bias

        return df.with_columns(pred.alias(name))


class ElasticNet:
    """
    Elastic Net Regression.
    """

    def __init__(
        self,
        l1_reg: float,
        l2_reg: float,
        has_bias: bool = False,
        tol: float = 1e-5,
        max_iter: int = 2000,
        feature_names_in_: List[str] | None = None,
    ):
        """
        Initializes an ElasticNet regressor. This is equivalent to Sklearn's Elastic Net if you set
        alpha and l1_ratio to be: `alpha = l1_reg + l2_reg`, and `l1_ratio = l1_reg / (l1_reg + l2_reg)`.

        Parameters
        ----------
        l1_reg
            The l1 regularization parameters for the elastic net.
        l2_reg
            The l2 regularization parameters for the elastic net.
        has_bias
            Whether to add a bias term. Also known as intercept in other packages.
        tol
            When updates are smaller than tol, the algorithm will stop.
        max_iter
            The max number of iterations the algorithm will run.
        feature_names_in_
            Names for the incoming features, if available. If None, the names will be empty. They will be
            learned if .fit_df() is run later, or .set_input_features() is set later.
        """
        if l1_reg <= 0.0 and l2_reg <= 0.0:
            raise ValueError("Cannot have both l1_reg and l2_reg <= 0.")

        self._en = PyElasticNet(l1_reg, l2_reg, has_bias, tol, max_iter)
        self.feature_names_in_: List[str] = (
            [] if feature_names_in_ is None else list(feature_names_in_)
        )

    @classmethod
    def from_values(
        cls, coeffs: List[float], bias: float = 0.0, feature_names_in_: List[str] | None = None
    ) -> Self:
        """
        Constructs a LR class instance from coefficients and bias values.

        Parameters
        ----------
        coeffs
            Iterable of numbers representing the coefficients
        bias
            Value for the bias term
        feature_names_in_
            Names for the incoming features, if available. If None, the names will be empty. They will be
            learned if .fit_df() is run later, or .set_input_features() is set later.
        """
        coefficients = np.ascontiguousarray(coeffs, dtype=np.float64).flatten()
        elastic_net = cls(
            float("nan"),
            float("nan"),
            has_bias=(bias != 0.0),
            tol=1e-5,
            max_iter=2000,
            feature_names_in_=feature_names_in_,
        )
        elastic_net._en.set_coeffs_and_bias(coefficients, bias)
        return elastic_net

    def is_fit(self) -> bool:
        return self._en.is_fit()

    def __repr__(self) -> str:
        output = f"Elastic Net Model\nl1, l2 regularizers: {self._en.regularizers}\n"
        if self._en.is_fit():
            output += f"Coefficients: {list(round(x, 5) for x in self._en.coeffs)}\n"
            output += f"Bias/Intercept: {self._en.bias}\n"
        else:
            output += "Not fitted yet."
        return output

    def set_input_features(self, features: List[str]) -> Self:
        """
        Sets the names of input features.

        Parameters
        ----------
        features
            List of strings.
        """
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self

    def coeffs(self) -> np.ndarray:
        """
        Returns a copy of the coefficients.
        """
        return self._en.coeffs

    def has_bias(self) -> bool:
        return self._en.has_bias()

    def bias(self) -> float:
        return self._en.bias

    def fit(self, X: np.ndarray, y: np.ndarray, null_policy: NullPolicy = "ignore") -> Self:
        """
        Fit the Elastic Net model on NumPy data.

        Parameters
        ----------
        X
            The feature Matrix. NumPy 2D matrix only.
        y
            The target data. NumPy array. Must be reshape-able to (-1, 1).
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.
        """
        self._en.fit(*_handle_nans_in_np(X, y.reshape((-1, 1)), null_policy))
        return self

    def fit_df(
        self,
        df: PolarsFrame,
        features: List[str],
        target: str,
        null_policy: NullPolicy = "skip",
    ) -> Self:
        """
        Fit the Elastic Net model on a dataframe. This will overwrite previously set feature names.
        The null policy only handles null values in df, not NaN values. It is the user's responsibility to handle
        NaN values if they exist in their pipeline.

        Parameters
        ----------
        df
            Either an eager or a lazy Polars dataframe.
        features
            List of strings of column names.
        target
            The target column's name.
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.
        """
        df2 = (
            _handle_nulls_in_df(df.lazy(), features, target, null_policy)
            .select(*features, target)
            .collect()
        )
        X = df2.select(features).to_numpy()
        y = df2.select(target).to_numpy()
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self.fit(X, y)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Returns the prediction of this linear model.

        Parameters
        ----------
        X
            Data to predict on, as a matrix
        """
        return self._en.predict(X).reshape((-1, 1))

    def predict_df(self, df: PolarsFrame, name: str = "prediction") -> PolarsFrame:
        """
        Computes the prediction of the linear model and append it as a column in the dataframe. If input
        is lazy, output will be lazy.

        Parameters
        ----------
        df
            Either an eager or a lazy Polars dataframe.
        name
            The name of the prediction column
        """
        if len(self.feature_names_in_) <= 0:
            raise ValueError(
                "The linear model is not fitted on a dataframe, or no feature names have been given."
                "Not enough info to predict on a dataframe. Hint: try .fit_df() or .set_input_features()."
            )

        pred = pl.sum_horizontal(
            beta * pl.col(c) for c, beta in zip(self.feature_names_in_, self._en.coeffs)
        )
        bias = self._en.bias
        if bias != 0.0:
            pred = pred + bias

        return df.with_columns(pred.alias(name))


class OnlineLR:
    """
    Normal or Ridge Online Regression. This doesn't support dataframe inputs.

    Because of implementation details, it is not recommended to set has_bias = True here
    if runtime speed is crucial.

    Null Behaviors:
    1. During the initial fit, no nulls/NaNs should be present
    2. During online updates, if the record has null/NaN, then it will be ignored. Nothing will be updated.
    """

    def __init__(
        self,
        lambda_: float = 0.0,
        has_bias: bool = False,
    ):
        """
        lambda_
            The L2 regularization factor
        has_bias
            Whether this should fit the bias term
        """
        self._lr = PyOnlineLR(lambda_, has_bias)

    @classmethod
    def from_coeffs_bias_inverse(cls, coeffs: List[float], bias: float, inv: np.ndarray) -> Self:
        """
        Constructs an online linear regression instance from coefficients, inverse. This copies
        data.

        Parameters
        ----------
        coeffs
            Iterable of numbers representing the coefficients
        bias
            The bias term
        inv
            2D NumPy matrix representing the inverse of XtX in a regression problem.
        """
        coefficients = np.ascontiguousarray(coeffs, dtype=np.float64).flatten()
        lr = cls(has_bias=(bias > 0.0), lambda_=0.0)
        lr._lr.set_coeffs_bias_inverse(coefficients, bias, inv)
        return lr

    def is_fit(self) -> bool:
        return self._lr.is_fit()

    def __repr__(self) -> str:
        if self._lr.lambda_ > 0.0:
            output = "Online Linear Regression (Ridge) Model\n"
        else:
            output = "Online Linear Regression Model\n"

        if self._lr.is_fit():
            output += f"Coefficients: {list(round(x, 5) for x in self._lr.coeffs)}\n"
            output += f"Bias/Intercept: {self._lr.bias}\n"
        else:
            output += "Not fitted yet."
        return output

    def coeffs(self) -> np.ndarray:
        """
        Returns a copy of the current coefficients.
        """
        return self._lr.coeffs

    def bias(self) -> float:
        self._lr.bias

    def inv(self) -> np.ndarray:
        """
        Returns a copy of the current inverse matrix (inverse of XtX in a linear regression).
        """
        return self._lr.inv

    def fit(self, X: np.ndarray, y: np.ndarray) -> Self:
        """
        Initial Fit for the online linear regression model on NumPy data.

        Parameters
        ----------
        X
            The feature Matrix. NumPy 2D matrix only.
        y
            The target data. NumPy array. Must be reshape-able to (-1, 1).
        """
        if np.any(np.isnan(X)) | np.any(np.isnan(y)):
            raise ValueError(
                "Online regression currently must fit without null for the initial fit."
            )

        self._lr.fit(X, y)
        return self

    def update(self, X: np.ndarray, y: np.ndarray | float, c: float = 1.0) -> Self:
        """
        Updates the online linear regression model with one row of data. If the row contains np.nan,
        it will be ignored.

        Parameters
        ----------
        X
            Either a a 1d array or a 2d array with 1 row. Must be reshapeable to a matrix with 1 row.
        y
            Either a scalar, or a 1d array with 1 element, or a 2d array of size 1x1.
        c
            The middle term (C) in the woodbury matrix identity. A value of 1.0 means we add
            the impact of the new data, and a value of -1.0 means we remove the impact of the
            data. Any other value will `scale` the impact of the data.
        """
        if not self.is_fit():
            raise ValueError("You cannot update before the initial fit of the matrix.")

        x_2d = X.reshape((1, -1))
        if isinstance(y, float):
            y_2d = np.array([[y]])
        else:
            y_2d = y.reshape((1, 1))

        self._lr.update(x_2d, y_2d, c)
        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Returns the prediction of this online linear model.

        Parameters
        ----------
        X
            Data to predict on, as a matrix
        """
        return self._lr.predict(X).reshape((-1, 1))


# ------------------------------------------------------------------------------------

GLMFamily: TypeAlias = Literal["gaussian", "normal", "poisson", "binomial", "logistic", "gamma"]
GLMSolver: TypeAlias = Literal["irls", "lbfgs"]
LinkFunction: TypeAlias = Literal["id", "identity", "log", "logit", "inverse"]


class GLM:
    """
    Generalized Linear Models.

    The GLM generalizes linear regression by allowing the linear model to be related to the response
    variable via a link function and by allowing the magnitude of the variance of each measurement to be a
    function of its predicted value.

    Currently, the variance function will be determined by the link function.
    If a family is given, then the canonical link function is used. Here is a mapping between currently
    implemented families and their link functions:

    gaussian / normal    ==>   id (x -> x)
    poisson              ==>   log (x -> ln(x))
    binomial / logistic  ==>   logit (x -> ln(x/(1-x)))
    gamma                ==>   inverse (x -> 1/x)

    Reference
    ---------
    https://en.wikipedia.org/wiki/Generalized_linear_model
    """

    def __init__(
        self,
        add_bias: bool = False,
        solver: GLMSolver = "irls",
        family: GLMFamily = "normal",
        max_iter: int = 100,
        tol: float = 1e-8,
        feature_names_in_: List[str] | None = None,
    ):
        # lambda_: float = 0.0,
        """
        Parameters
        ----------
        family
            One of "gaussian", "normal", "poisson", "binomial", "logistic", "gamma". Note "gaussian" and
            "normal" represent the same family.
        add_bias
            Whether to add a bias term. Also known as intercept in other packages.
        max_iter
            Max number of iterations for the algorithm
        tol
            The tolerance for convergence
        feature_names_in_
            Names for the incoming features, if available. If None, the names will be empty. They will be
            learned if .fit_df() is run later, or .set_input_features() is set later.
        """
        if solver not in ["irls"]:
            raise NotImplementedError

        if max_iter < 1:
            raise ValueError("`max_iter` must be > 1.")

        if family not in ["gaussian", "normal", "poisson", "binomial", "logistic", "gamma"]:
            raise NotImplementedError

        self._glm = PyGLM(
            add_bias=add_bias, family=family, solver=solver, max_iter=max_iter, tol=abs(tol)
        )
        self.feature_names_in_: List[str] = (
            [] if feature_names_in_ is None else list(feature_names_in_)
        )

    def __repr__(self) -> str:
        """
        Shows a textual representation of the GLM.
        """
        return self._glm.describe()

    # @classmethod
    # def from_values(
    #     cls,
    #     coeffs: List[float],
    #     link: LinkFunction,
    #     bias: float = 0.0,
    #     feature_names_in_: List[str] | None = None
    # ) -> Self:
    #     """
    #     Constructs a LR class instance from coefficients and bias values.

    #     Parameters
    #     ----------
    #     coeffs
    #         Iterable of numbers representing the coefficients
    #     link
    #         One of ["id", "log", "logit", "inverse"].
    #     bias
    #         Value for the bias term
    #     feature_names_in_
    #         Names for the incoming features, if available. If None, the names will be empty. They will be
    #         learned if .fit_df() is run later, or .set_input_features() is set later.
    #     """
    #     coefficients = np.ascontiguousarray(coeffs, dtype=np.float64).flatten()
    #     lr = cls(
    #         add_bias=(bias != 0.0),
    #         solver="irls",
    #         feature_names_in_=feature_names_in_,
    #     )
    #     lr._lr.set_coeffs_and_bias(coefficients, bias)
    #     return lr

    def is_fit(self) -> bool:
        return self._glm.is_fit()

    # def __repr__(self) -> str:
    #     if self._lr.lambda_ > 0.0:
    #         output = "Linear Regression (Ridge) Model\n"
    #     else:
    #         output = "Linear Regression Model\n"

    #     if self._lr.is_fit():
    #         output += f"Coefficients: {list(round(x, 5) for x in self._lr.coeffs)}\n"
    #         output += f"Bias/Intercept: {self._lr.bias}\n"
    #     else:
    #         output += "Not fitted yet."
    #     return output

    def set_input_features(self, features: List[str]) -> Self:
        """
        Sets the names of input features.

        Parameters
        ----------
        features
            List of strings.
        """
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self

    def coeffs(self) -> np.ndarray:
        """
        Returns a copy of the coefficients.
        """
        return self._glm.coeffs

    def bias(self) -> float:
        return self._glm.bias

    def fit(self, X: np.ndarray, y: np.ndarray, null_policy: NullPolicy = "ignore") -> Self:
        """
        Fit the GLM model on NumPy data.

        Parameters
        ----------
        X
            The feature Matrix. NumPy 2D matrix only.
        y
            The target data. NumPy array. Must be reshape-able to (-1, 1).
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.
        """
        self._glm.fit(*_handle_nans_in_np(X, y.astype(np.float64).reshape((-1, 1)), null_policy))
        return self

    def fit_df(
        self,
        df: PolarsFrame,
        features: List[str],
        target: str,
        null_policy: NullPolicy = "skip",
        show_report: bool = False,
    ) -> Self:
        """
        Fit the GLM model on a dataframe. This will overwrite previously set feature names.
        The null policy only handles null values in df, not NaN values. It is the user's responsibility to handle
        NaN values if they exist in their pipeline.

        Parameters
        ----------
        df
            Either an eager or a lazy Polars dataframe.
        features
            List of strings of column names.
        target
            The target column's name.
        null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
            One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
            fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
            the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
            columns. If target has null, then the row will still be dropped.

        """
        df2 = (
            _handle_nulls_in_df(df.lazy(), features, target, null_policy)
            .select(*features, target)
            .collect()
        )
        X = df2.select(features).to_numpy()
        y = df2.select(target).to_numpy()
        self.feature_names_in_.clear()
        self.feature_names_in_ = list(features)
        return self.fit(X, y)

    def predict(self, X: np.ndarray, linear: bool = False) -> np.ndarray:
        """
        Returns the prediction of this linear model.

        Parameters
        ----------
        X
            Data to predict on, as a matrix
        linear
            If true, return the linear predictor eta instead of the expected value of
            the response variable, E[Y|X].
        """
        return self._glm.predict(X, linear).reshape((-1, 1))

    # def predict_df(self, df: PolarsFrame, name: str = "prediction") -> PolarsFrame:
    #     """
    #     Computes the prediction of the linear model and append it as a column in the dataframe. If input
    #     is lazy, output will be lazy.

    #     Parameters
    #     ----------
    #     df
    #         Either an eager or a lazy Polars dataframe.
    #     name
    #         The name of the prediction column
    #     """
    #     if len(self.feature_names_in_) <= 0:
    #         raise ValueError(
    #             "The linear model is not fitted on a dataframe, or no feature names have been given."
    #             "Not enough info to predict on a dataframe. Hint: try .fit_df() or .set_input_features()."
    #         )

    #     pred = pl.sum_horizontal(
    #         beta * pl.col(c) for c, beta in zip(self.feature_names_in_, self._lr.coeffs)
    #     )
    #     bias = self._lr.bias
    #     if bias != 0.0:
    #         pred = pred + bias

    #     return df.with_columns(pred.alias(name))



import functools
import timeit
import unicodedata
from pathlib import Path
from typing import Callable

import polars as pl
import polars_ds as pds

BASE_PATH = Path(__file__).resolve().parents[0]

TIMING_RUNS = 10


class Bench:
    def __init__(
        self,
        df: pl.DataFrame,
        sizes: list[int] = [100, 1_000, 10_000, 100_000, 1_000_000],
        timing_runs: int = 10,
    ):
        self.benchmark_data = {"Function": [], "Size": [], "Time": []}
        self.df = df
        self.sizes = sizes
        self.timing_runs = timing_runs

    def run(self, funcs: list[Callable]):
        for n_rows in self.sizes:
            df = self.df.sample(n_rows, seed=208)

            for f in funcs:
                func_name = f.func.__name__ if isinstance(f, functools.partial) else f.__name__
                time = timeit.timeit(lambda: f(df), number=self.timing_runs)

                self.benchmark_data["Function"].append(func_name)
                self.benchmark_data["Size"].append(n_rows)
                self.benchmark_data["Time"].append(time)

        return self

    def save(self, file: Path):
        pl.DataFrame(self.benchmark_data).write_parquet(file)


def python_remove_non_ascii(df: pl.DataFrame):
    df.select(
        pl.col("RANDOM_STRING").map_elements(
            lambda s: s.encode("ascii", errors="ignore").decode(),
            return_dtype=pl.String,
        )
    )


def regex_remove_non_ascii(df: pl.DataFrame):
    df.select(pl.col("RANDOM_STRING").str.replace_all(r"[^\p{Ascii}]", ""))


def pds_remove_non_ascii(df: pl.DataFrame):
    df.select(pds.replace_non_ascii("RANDOM_STRING"))


def python_remove_diacritics(df: pl.DataFrame):
    df.select(
        pl.col("RANDOM_STRING").map_elements(
            lambda s: unicodedata.normalize("NFD", s).encode("ASCII", "ignore"),
            return_dtype=pl.String,
        )
    )


def pds_remove_diacritics(df: pl.DataFrame):
    df.select(pds.remove_diacritics("RANDOM_STRING"))


def python_normalize_string(df: pl.DataFrame):
    df.select(
        pl.col("RANDOM_STRING").map_elements(
            lambda s: unicodedata.normalize("NFD", s), return_dtype=pl.String
        )
    )


def pds_normalize_string(df: pl.DataFrame):
    df.select(pds.normalize_string("RANDOM_STRING", "NFD"))


def python_map_words(df: pl.DataFrame, mapping: dict[str, str]):
    df.select(
        pl.col("RANDOM_ADDRESS").map_elements(
            lambda s: " ".join(mapping.get(word, word) for word in s.split()),
            return_dtype=pl.String,
        )
    )


def regex_map_words(df: pl.DataFrame, mapping: dict[str, str]):
    expr = pl.col("RANDOM_ADDRESS")
    for k, v in mapping.items():
        expr = expr.str.replace_all(k, v)
    df.select(expr)


def pds_map_words(df: pl.DataFrame, mapping: dict[str, str]):
    df.select(pds.map_words("RANDOM_ADDRESS", mapping))


def python_normalize_whitespace(df: pl.DataFrame):
    df.select(
        pl.col("RANDOM_STRING").map_elements(lambda s: " ".join(s.split()), return_dtype=pl.String)
    )


def python_normalize_whitespace_only_spaces(df: pl.DataFrame):
    df.select(
        pl.col("RANDOM_STRING").map_elements(
            lambda s: " ".join(s.split(" ")), return_dtype=pl.String
        )
    )


def expr_normalize_whitespace_only_spaces(df: pl.DataFrame):
    df.select(pl.col("RANDOM_STRING").str.split(" ").list.join(" "))


def pds_normalize_whitespace(df: pl.DataFrame):
    df.select(pds.normalize_whitespace("RANDOM_STRING"))


def pds_normalize_whitespace_only_spaces(df: pl.DataFrame):
    df.select(pds.normalize_whitespace("RANDOM_STRING", only_spaces=True))


def main():
    benchmark_df = pl.read_parquet(BASE_PATH / "benchmark_df.parquet")

    map_words_mapping = {
        "Apt.": "Apartment",
        "NY": "New York",
        "CT": "Connecticut",
        "Street": "ST",
        "Bypass": "BYP",
        "GA": "Georgia",
        "Parkways": "Pkwy",
        "PA": "Pennsylvania",
    }

    Bench(benchmark_df).run(
        [
            python_remove_non_ascii,
            regex_remove_non_ascii,
            pds_remove_non_ascii,
            python_remove_diacritics,
            pds_remove_diacritics,
            python_normalize_string,
            pds_normalize_string,
            functools.partial(python_map_words, mapping=map_words_mapping),
            functools.partial(
                regex_map_words,
                mapping={f"\b{k}\b": v for k, v in map_words_mapping.items()},
            ),
            functools.partial(pds_map_words, mapping=map_words_mapping),
            python_normalize_whitespace,
            python_normalize_whitespace_only_spaces,
            expr_normalize_whitespace_only_spaces,
            pds_normalize_whitespace,
            pds_normalize_whitespace_only_spaces,
        ]
    ).save(BASE_PATH / "benchmark_data.parquet")


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest
import polars as pl
import numpy as np
import polars_ds as pds
from polars.testing import assert_frame_equal, assert_series_equal


def test_mcc():
    from sklearn.metrics import matthews_corrcoef

    df = (
        pds.frame(size=2000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
        )
        .with_columns(
            pl.col("x1").round(),
            pl.col("x2").round(),
        )
    )

    sklearn_result = matthews_corrcoef(df["x1"], df["x2"])
    result = df.select(pds.query_mcc("x1", "x2")).item(0, 0)
    assert np.isclose(result, sklearn_result)


def test_pca():
    from sklearn.decomposition import PCA

    df = pds.frame(size=2000).select(
        pds.random(0.0, 1.0).alias("x1"),
        pds.random(0.0, 1.0).alias("x2"),
        pds.random(0.0, 1.0).alias("x3"),
    )

    singular_values = df.select(pds.singular_values("x1", "x2", "x3").alias("res"))["res"][
        0
    ].to_numpy()

    pca = PCA()
    data_matrix = df.select("x1", "x2", "x3").to_numpy().astype(np.float64)
    pca.fit(data_matrix)
    ans_singular_values = pca.singular_values_

    assert np.isclose(singular_values, ans_singular_values).all()

    singular_values = df.select(pds.singular_values("x1", "x2", "x3").alias("res"))["res"][
        0
    ].to_numpy()

    vectors = df.select(pds.pca("x1", "x2", "x3").alias("vectors")).unnest("vectors")[
        "weight_vector"
    ]

    ans_vectors = pca.components_
    for i in range(len(vectors)):
        vi = vectors[i].to_numpy()
        ans_vi = ans_vectors[i, :]
        # The principal vector can be either v or -v. It doesn't matter and depends on backend algo's choice.
        # Have a more relaxed rtol
        assert np.isclose(np.abs(vi), np.abs(ans_vi), rtol=1e-5).all()


def test_copula_entropy():
    from numpy.random import multivariate_normal as mnorm
    import copent

    rho = 0.6
    mean1 = [0, 0]
    cov1 = [[1, rho], [rho, 1]]
    x = mnorm(mean1, cov1, 200)  # bivariate gaussian
    ce1 = copent.copent(x, dtype="euclidean")  # estimated copula entropy

    df = pl.from_numpy(x, schema=["x1", "x2"])
    res = df.select(pds.query_copula_entropy("x1", "x2", k=3)).item(0, 0)

    assert np.isclose(res, ce1)


def test_cond_indep_and_transfer():
    from copent import ci, transent

    df = pds.frame(size=2_000).select(
        pds.random(0.0, 1.0).alias("x1"),
        pds.random(0.0, 1.0).alias("x2"),
        pds.random(0.0, 1.0).alias("x3"),
    )

    ci_ans = ci(df["x1"].to_numpy(), df["x2"].to_numpy(), df["x3"].to_numpy(), dtype="euclidean")
    ci_res = df.select(pds.query_cond_indep("x1", "x2", "x3", k=3)).item(0, 0)

    assert np.isclose(ci_ans, ci_res)

    t_ans = transent(df["x1"].to_numpy(), df["x2"].to_numpy(), dtype="euclidean")
    t_res = df.select(pds.query_transfer_entropy("x1", "x2", k=3)).item(0, 0)

    assert np.isclose(t_ans, t_res)


def test_xi_corr():
    df = pds.frame(size=2_000).select(
        pds.random(0.0, 12.0).alias("x"),
        pds.random(0.0, 1.0).alias("y"),
    )

    from xicor.xicor import Xi

    x = df["x"].to_numpy()
    y = df["y"].to_numpy()
    xi_obj = Xi(x, y)
    ans_statistic = xi_obj.correlation
    test_statistic = df.select(pds.xi_corr("x", "y")).item(0, 0)

    assert np.isclose(ans_statistic, test_statistic, rtol=1e-5)


def test_bicor():
    df = pds.frame(size=2_000).select(
        pds.random(0.0, 1.0).alias("x"),
        pds.random(0.0, 1.0).alias("y"),
    )

    from astropy.stats import biweight_midcorrelation

    x = df["x"].to_numpy()
    y = df["y"].to_numpy()
    answer = biweight_midcorrelation(x, y)
    test_result = df.select(pds.bicor("x", "y")).item(0, 0)

    assert np.isclose(answer, test_result)


def test_kendall_tau():
    from scipy.stats import kendalltau

    df = pds.frame(size=2000).select(
        pds.random_int(0, 200).alias("x"),
        pds.random_int(0, 200).alias("y"),
    )

    test = df.select(pds.kendall_tau("x", "y")).item(0, 0)

    res = kendalltau(df["x"].to_numpy(), df["y"].to_numpy())

    assert np.isclose(test, res.statistic)


@pytest.mark.parametrize(
    "a, value, res",
    [
        ([1, 2, 3, 4, 5, None], 2, 4),
        ([1, 2, 3, 4, 5, None], 6, 0),
    ],
)
def test_longest_streak(a, value, res):
    # >=
    df = pl.DataFrame({"a": a})
    longest = df.select(pds.query_longest_streak(pl.col("a") >= value)).item(0, 0)
    assert longest == res


@pytest.mark.parametrize(
    "a, value, res",
    [
        ([1, 2, 3, 4, 5, None], 2, 2),
        ([1, 2, 3, 4, 5, None], 6, 5),  # None doesn't count
    ],
)
def test_longest_streak_2(a, value, res):
    # <=
    df = pl.DataFrame({"a": a})
    longest = df.select(pds.query_longest_streak(pl.col("a") <= value)).item(0, 0)
    assert longest == res


@pytest.mark.parametrize(
    "df, kernel, res_full, res_valid, res_same",
    [
        (
            pl.DataFrame({"a": [5, 6, 7, 8, 9]}),
            [1, 0, -1],
            pl.DataFrame({"a": pl.Series([5, 6, 2, 2, 2, -8, -9], dtype=pl.Float64)}),
            pl.DataFrame({"a": pl.Series([2, 2, 2], dtype=pl.Float64)}),
            pl.DataFrame({"a": pl.Series([6, 2, 2, 2, -8], dtype=pl.Float64)}),
        ),
    ],
)
def test_convolve(df, kernel, res_full, res_valid, res_same):
    res = df.select(pds.convolve("a", kernel, mode="full"))

    assert_frame_equal(res, res_full)

    res = df.select(pds.convolve("a", kernel, mode="valid"))

    assert_frame_equal(res, res_valid)

    res = df.select(pds.convolve("a", kernel, mode="same"))

    assert_frame_equal(res, res_same)

    res = df.select(pds.convolve("a", kernel, mode="full", parallel=True))

    assert_frame_equal(res, res_full)

    res = df.select(pds.convolve("a", kernel, mode="valid", parallel=True))

    assert_frame_equal(res, res_valid)

    res = df.select(pds.convolve("a", kernel, mode="same", parallel=True))

    assert_frame_equal(res, res_same)

    res = df.select(pds.convolve("a", kernel, mode="full", method="fft"))

    assert_frame_equal(res, res_full)

    res = df.select(pds.convolve("a", kernel, mode="valid", method="fft"))

    assert_frame_equal(res, res_valid)

    res = df.select(pds.convolve("a", kernel, mode="same", method="fft"))

    assert_frame_equal(res, res_same)


@pytest.mark.parametrize(
    "arr, n",
    [
        # n is Optional[int]
        (np.random.normal(size=200), None),
        (np.random.normal(size=200), 100),
    ],
)
def test_fft(arr, n):
    import scipy as sp

    df = pl.DataFrame({"a": arr})
    res = df.select(pds.rfft("a", n=n).alias("fft")).select(
        pl.col("fft").arr.first().alias("re"), pl.col("fft").arr.last().alias("im")
    )
    real_test = res["re"].to_numpy()
    im_test = res["im"].to_numpy()

    ans = np.fft.rfft(arr, n=n)
    real = ans.real
    imag = ans.imag
    assert np.isclose(real_test, real).all()
    assert np.isclose(im_test, imag).all()

    # Always run a check against scipy, with return_full=True as well
    res2 = df.select(pds.rfft("a", return_full=True).alias("fft")).select(
        pl.col("fft").arr.first().alias("re"), pl.col("fft").arr.last().alias("im")
    )
    real_test = res2["re"].to_numpy()
    im_test = res2["im"].to_numpy()
    ans = sp.fft.fft(arr)  # always full fft
    real = ans.real
    imag = ans.imag
    assert np.isclose(real_test, real).all()
    assert np.isclose(im_test, imag).all()


@pytest.mark.parametrize(
    "df",
    [
        (
            pl.DataFrame(
                {
                    "target": np.random.randint(0, 3, size=1_000),
                    "a": np.random.normal(size=1_000),
                }
            )
        ),
    ],
)
def test_f_test(df):
    from sklearn.feature_selection import f_classif

    res = df.select(pds.f_test(pl.col("a"), group=pl.col("target")))
    res = res.item(0, 0)  # A dictionary
    statistic = res["statistic"]
    pvalue = res["pvalue"]

    scikit_res = f_classif(df["a"].to_numpy().reshape(-1, 1), df["target"].to_numpy())
    scikit_s = scikit_res[0][0]
    scikit_p = scikit_res[1][0]

    assert np.isclose(statistic, scikit_s)
    assert np.isclose(pvalue, scikit_p)


@pytest.mark.parametrize(
    "df",
    [
        (
            pl.DataFrame(
                {
                    "x1": np.random.normal(size=1_000),
                    "x2": np.random.normal(size=1_000),
                }
            )
        ),
    ],
)
def test_mann_whitney_u(df):
    from scipy.stats import mannwhitneyu

    res = df.select(pds.mann_whitney_u("x1", "x2"))
    res = res.item(0, 0)  # A dictionary
    res_statistic = res["statistic"]
    res_pvalue = res["pvalue"]
    answer = mannwhitneyu(df["x1"].to_numpy(), df["x2"].to_numpy())
    assert np.isclose(res_statistic, answer.statistic)
    assert np.isclose(res_pvalue, answer.pvalue)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": list(range(24))}),
            [11, 5, 1, 1, 1, 1, 1, 1, 1],
        ),
        (
            pl.DataFrame({"a": [1.0, 2.0, 3.0, 4.0, float("nan"), float("inf"), None]}),
            [1, 1, 1, 1, 0, 0, 0, 0, 0],  # NaN, Inf, None are ignored
        ),
    ],
)
def test_first_digit_cnt(df, res):
    assert_frame_equal(
        df.select(pds.query_first_digit_cnt("a").explode().cast(pl.UInt32)),
        pl.DataFrame({"a": pl.Series(values=res, dtype=pl.UInt32)}),
    )
    assert_frame_equal(
        df.lazy().select(pds.query_first_digit_cnt("a").explode().cast(pl.UInt32)).collect(),
        pl.DataFrame({"a": pl.Series(values=res, dtype=pl.UInt32)}),
    )


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": [2.123, None, -2.111, float("nan")]}),
            pl.DataFrame({"a": [2.0, None, -2.0, float("nan")]}),
        ),
    ],
)
def test_trunc(df, res):
    assert_frame_equal(df.select(pds.trunc("a")), res)
    assert_frame_equal(df.lazy().select(pds.trunc("a")).collect(), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": [2.123, None, -2.111, float("nan")]}),
            pl.DataFrame({"a": [0.123, None, -0.111, float("nan")]}),
        ),
    ],
)
def test_fract(df, res):
    assert_frame_equal(df.select(pds.fract("a")), res)
    assert_frame_equal(df.lazy().select(pds.fract("a")).collect(), res)


@pytest.mark.parametrize(
    "df, other, res",
    [
        (
            pl.DataFrame({"a": [1, 2, 3, 4, 5], "b": [1, 2, 2, 2, 10]}),
            3,
            pl.DataFrame({"a": [1, 1, 3, 1, 1]}),
        ),
        (
            pl.DataFrame({"a": [1, 2, 3, 4, 5], "b": [1, 2, 2, 2, 10]}),
            pl.col("b"),
            pl.DataFrame({"a": [1, 2, 1, 2, 5]}),
        ),
    ],
)
def test_gcd(df, other, res):
    assert_frame_equal(df.select(pds.gcd("a", other).cast(pl.Int64)), res)

    assert_frame_equal(df.lazy().select(pds.gcd("a", other).cast(pl.Int64)).collect(), res)


@pytest.mark.parametrize(
    "df, other, res",
    [
        (
            pl.DataFrame({"a": [1, 2, 3, 4, 5], "b": [1, 2, 2, 2, 10]}),
            3,
            pl.DataFrame({"a": [3, 6, 3, 12, 15]}),
        ),
        (
            pl.DataFrame({"a": [1, 2, 3, 4, 5], "b": [1, 2, 2, 2, 10]}),
            pl.col("b"),
            pl.DataFrame({"a": [1, 2, 6, 4, 10]}),
        ),
    ],
)
def test_lcm(df, other, res):
    assert_frame_equal(df.select(pds.lcm("a", other).cast(pl.Int64)), res)

    assert_frame_equal(df.lazy().select(pds.lcm("a", other).cast(pl.Int64)).collect(), res)


@pytest.mark.parametrize(
    "df, x, res",
    [
        (
            pl.DataFrame({"a": [1, 2, 3]}),
            0.1,
            pl.DataFrame({"a": [0.4]}),
        ),
    ],
)
def test_trapz(df, x, res):
    assert_frame_equal(df.select(pds.integrate_trapz("a", x=x)), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame(
                {"y": [1, 0, 1, 1, 1, 0, 0, 1], "a": ["a", "b", "c", "a", "b", "c", "a", "a"]}
            ),
            pl.DataFrame({"y": [0.6277411625893767]}),
        ),
        (
            pl.DataFrame({"y": [1] * 8, "a": ["a", "b", "c", "a", "b", "c", "a", "a"]}),
            pl.DataFrame({"y": [-0.0]}),
        ),
    ],
)
def test_cond_entropy(df, res):
    assert_frame_equal(df.select(pds.query_cond_entropy("y", "a")), res)

    assert_frame_equal(df.lazy().select(pds.query_cond_entropy("y", "a")).collect(), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame(
                {
                    "y": [0, 1, 2, 0, 1],
                    "pred": [
                        [0.1, 0.5, 0.4],
                        [0.2, 0.6, 0.2],
                        [0.4, 0.1, 0.5],
                        [0.9, 0.05, 0.05],
                        [0.2, 0.5, 0.3],
                    ],
                }
            ),
            pl.DataFrame({"a": [0.8610131187075506]}),
        ),
    ],
)
def test_cross_entropy(df, res):
    assert_frame_equal(df.select(pds.query_cat_cross_entropy("y", "pred").alias("a")), res)

    assert_frame_equal(
        df.lazy().select(pds.query_cat_cross_entropy("y", "pred").alias("a")).collect(),
        res,
    )


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": [[1, 2, 3], [2, 3]], "b": [[1, 3], [1]]}),
            pl.DataFrame({"res": [2 / 3, 0.0]}),
        ),
        (
            pl.DataFrame({"a": [["a", "b", "c"], ["b", "c"]], "b": [["a", "b"], ["c"]]}),
            pl.DataFrame({"res": [2 / 3, 0.5]}),
        ),
    ],
)
def test_jaccard_row(df, res):
    assert_frame_equal(df.select(pds.jaccard_row("a", "b").alias("res")), res)

    assert_frame_equal(df.lazy().select(pds.jaccard_row("a", "b").alias("res")).collect(), res)


def test_lin_reg_against_sklearn():
    # Random data + noise
    df = (
        pds.frame(size=5_000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5 + pl.col("x2") * 0.1 - pl.col("x3") * 0.15 + pds.random() * 0.0001
        )
    )

    #
    from sklearn import linear_model

    x = df.select("x1", "x2", "x3").to_numpy()
    y = df["y"].to_numpy()

    # sklearn, normal, with bias
    reg = linear_model.LinearRegression(fit_intercept=True)
    reg.fit(x, y)

    # pds, normal, with bias
    normal_coeffs = df.select(
        pds.lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            add_bias=True,
        ).alias("pred")
    ).explode("pred")
    all_coeffs = normal_coeffs["pred"].to_numpy()
    # non-bias terms, some precision differences expected.
    assert np.all(np.abs(all_coeffs[:3] - reg.coef_) < 1e-5)
    assert np.isclose(all_coeffs[-1], reg.intercept_, rtol=1e-5)

    # sklearn, L2 (Ridge), with bias
    reg = linear_model.Ridge(alpha=0.1, fit_intercept=True)
    reg.fit(x, y)

    # pds, normal, with bias
    normal_coeffs = df.select(
        pds.lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            l2_reg=0.1,
            add_bias=True,
        ).alias("pred")
    ).explode("pred")
    all_coeffs = normal_coeffs["pred"].to_numpy()
    coeffs = all_coeffs[:3]
    bias = all_coeffs[-1]
    # non-bias terms, slightly bigger precision differences expected.
    assert np.all(np.abs(coeffs - reg.coef_) < 1e-3)
    assert abs(bias - reg.intercept_) < 1e-3


def test_lasso_regression():
    # These tests have bigger precision tolerance because of different stopping criterions

    from sklearn import linear_model

    df = (
        pds.frame(size=5_000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5 + pl.col("x2") * 0.25 - pl.col("x3") * 0.15 + pds.random() * 0.0001
        )
    )

    x = df.select("x1", "x2", "x3").to_numpy()
    y = df["y"].to_numpy()

    for lambda_ in [0.01, 0.05, 0.1, 0.2]:
        df_res = df.select(
            pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=False).alias(
                "coeffs"
            )
        ).explode("coeffs")

        res = df_res["coeffs"].to_numpy()

        sklearn = linear_model.Lasso(alpha=lambda_, fit_intercept=False)
        sklearn.fit(x, y)
        res_sklearn = np.asarray(sklearn.coef_)
        assert np.all(np.abs(res_sklearn - res) < 1e-4)

    for lambda_ in [0.01, 0.05, 0.1, 0.2]:
        df_res = df.select(
            pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=lambda_, add_bias=True).alias("coeffs")
        ).explode("coeffs")

        res = df_res["coeffs"].to_numpy()
        res_coef = res[:3]
        res_bias = res[-1]

        sklearn = linear_model.Lasso(alpha=lambda_, fit_intercept=True)
        sklearn.fit(x, y)
        res_sklearn = np.asarray(sklearn.coef_)
        assert np.all(np.abs(res_sklearn - res_coef) < 1e-4)
        assert abs(res_bias - sklearn.intercept_) < 1e-4


def test_positive_lin_reg():
    #
    from sklearn.linear_model import LinearRegression, ElasticNet

    df = (
        pds.frame(size=5_000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            + pl.col("x3") * -0.15
            + pds.random() * 0.0001
        )
    )

    # Third coefficient should be 0 because this is non neg
    for bias in [True, False]:
        pds_result = df.select(
            pds.lin_reg(*[f"x{i + 1}" for i in range(3)], target="y", positive=True, add_bias=bias)
        ).item(0, 0)
        pds_result = pds_result.to_numpy()

        if not bias:
            assert np.all(pds_result >= 0.0)
        else:
            # Bias term can be negative
            assert np.all(pds_result[:-1] >= 0.0)

        reg_nnls = LinearRegression(positive=True, fit_intercept=bias)
        reg_nnls.fit(df.select("x1", "x2", "x3").to_numpy(), df["y"].to_numpy())

        if not bias:
            assert np.all(np.isclose(pds_result, reg_nnls.coef_, atol=1e-5))
        else:
            assert np.all(np.isclose(pds_result[:-1], reg_nnls.coef_, atol=1e-5))
            assert np.isclose(float(pds_result[-1]), reg_nnls.intercept_, atol=1e-5)

    for reg, bias in zip([0.01, 0.05, 0.1, 0.2], [False, True, False, True]):
        l1_reg = reg
        l2_reg = reg

        pds_result = df.select(
            pds.lin_reg(
                "x1", "x2", "x3", target="y", l1_reg=l1_reg, l2_reg=l2_reg, add_bias=bias
            ).alias("coeffs")
        ).item(0, 0)

        pds_result = pds_result.to_numpy()
        if not bias:
            assert np.all(pds_result >= 0.0)
        else:
            # Bias term can be negative
            assert np.all(pds_result[:-1] >= 0.0)

        alpha = l1_reg + l2_reg
        l1_ratio = l1_reg / (l1_reg + l2_reg)

        reg_nnls = ElasticNet(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=bias)
        reg_nnls.fit(df.select("x1", "x2", "x3").to_numpy(), df["y"].to_numpy())
        if not bias:  # allow for a higher error tolerance because of convergence differences
            assert np.all(np.isclose(pds_result, reg_nnls.coef_, atol=1e-4))
        else:
            assert np.all(np.isclose(pds_result[:-1], reg_nnls.coef_, atol=1e-4))
            assert np.isclose(float(pds_result[-1]), reg_nnls.intercept_, atol=1e-4)


def test_elastic_net_regression():
    # These tests have bigger precision tolerance because of different stopping criterions

    from sklearn import linear_model

    df = (
        pds.frame(size=5_000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5 + pl.col("x2") * 0.25 - pl.col("x3") * 0.15 + pds.random() * 0.0001
        )
    )

    x = df.select("x1", "x2", "x3").to_numpy()
    y = df["y"].to_numpy()

    for reg in [0.01, 0.05, 0.1, 0.2]:
        l1_reg = reg
        l2_reg = reg

        df_res = df.select(
            pds.lin_reg(
                "x1", "x2", "x3", target="y", l1_reg=l1_reg, l2_reg=l2_reg, add_bias=False
            ).alias("coeffs")
        ).explode("coeffs")

        res = df_res["coeffs"].to_numpy()

        alpha = l1_reg + l2_reg
        l1_ratio = l1_reg / (l1_reg + l2_reg)

        sklearn = linear_model.ElasticNet(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False)
        sklearn.fit(x, y)
        res_sklearn = np.asarray(sklearn.coef_)
        assert np.all(np.abs(res_sklearn - res) < 1e-4)


def test_recursive_lin_reg():
    # Test against the lstsq method with a fit whenver a new row is in the data
    size = 1_000
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            - pl.col("x3") * 0.15
            + pds.random() * 0.0001,
        )
    )

    start_with = 3

    df_recursive_lr = df.select(
        "y",
        pds.recursive_lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            start_with=start_with,
        ).alias("result"),
    ).unnest("result")

    for i in range(start_with, 30):
        coefficients = df.limit(i).select(
            pds.lin_reg(
                "x1",
                "x2",
                "x3",
                target="y",
            ).alias("coeffs")
        )["coeffs"]  # One element series

        normal_result = coefficients[0].to_numpy()
        # i - 1. E.g. use 3 rows of data to train, the data will be at row 2.
        recursive_result = df_recursive_lr["coeffs"][i - 1].to_numpy()
        assert np.all(np.abs(normal_result - recursive_result) < 1e-5)


def test_recursive_ridge():
    # Test against the lstsq method with a fit whenver a new row is in the data
    size = 1_000
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            - pl.col("x3") * 0.15
            + pds.random() * 0.0001,
        )
    )

    start_with = 3

    df_recursive_lr = df.select(
        "y",
        pds.recursive_lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            l2_reg=0.1,
            start_with=start_with,
        ).alias("result"),
    ).unnest("result")

    for i in range(start_with, 30):
        coefficients = df.limit(i).select(
            pds.lin_reg(
                "x1",
                "x2",
                "x3",
                target="y",
                l2_reg=0.1,
            ).alias("coeffs")
        )["coeffs"]  # One element series

        normal_result = coefficients[0].to_numpy()
        # i - 1. E.g. use 3 rows of data to train, the data will be at row 2.
        recursive_result = df_recursive_lr["coeffs"][i - 1].to_numpy()
        assert np.all(np.abs(normal_result - recursive_result) < 1e-5)


def test_rolling_lin_reg():
    # Test rolling lstsq by comparing it with a manually rolled lstsq result.
    # Test on multiple window sizes
    size = 500
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
            pl.Series(name="id", values=list(range(size))),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            - pl.col("x3") * 0.15
            + pds.random() * 0.0001,
        )
    )

    for window_size in [5, 8, 12, 15]:
        df_to_test = df.select(
            "id",
            "y",
            pds.rolling_lin_reg(
                "x1",
                "x2",
                "x3",
                target="y",
                window_size=window_size,
            ).alias("result"),
        ).unnest("result")  # .limit(10)
        df_to_test = df_to_test.filter(pl.col("id") >= window_size - 1).select("coeffs")

        results = []
        for i in range(len(df) - window_size + 1):
            temp = df.slice(i, length=window_size)
            results.append(temp.select(pds.lin_reg("x1", "x2", "x3", target="y").alias("coeffs")))

        df_answer = pl.concat(results)
        assert_frame_equal(df_to_test, df_answer)


# This only tests that nulls are correctly skipped.
def test_rolling_null_skips():
    size = 1000
    # Data with random nulls
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_columns(
            pl.when(pds.random() < 0.15).then(None).otherwise(pl.col("x1")).alias("x1"),
            pl.when(pds.random() < 0.15).then(None).otherwise(pl.col("x2")).alias("x2"),
            pl.when(pds.random() < 0.15).then(None).otherwise(pl.col("x3")).alias("x3"),
        )
        .with_columns(
            null_ref=pl.any_horizontal(
                pl.col("x1").is_null(), pl.col("x2").is_null(), pl.col("x3").is_null()
            ),
            y=pl.col("x1") * 0.15 + pl.col("x2") * 0.3 - pl.col("x3") * 1.5 + pds.random() * 0.0001,
        )
    )

    window_size = 6
    min_valid_rows = 5

    result = df.with_columns(
        pds.rolling_lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            window_size=window_size,
            min_valid_rows=min_valid_rows,
            null_policy="skip",
        ).alias("test")
    ).with_columns(
        pl.col("test").struct.field("coeffs").alias("coeffs"),
        pl.col("test").struct.field("coeffs").is_null().alias("is_null"),
    )

    nulls = df["null_ref"].to_list()  # list of bools
    rolling_should_be_null = [True] * (window_size - 1)
    for i in range(0, len(nulls) - window_size + 1):
        lower = i
        upper = i + window_size
        window_valid_count = window_size - np.sum(nulls[lower:upper])  # size - null count
        rolling_should_be_null.append((window_valid_count < min_valid_rows))

    answer = pl.Series(name="is_null", values=rolling_should_be_null)
    assert_series_equal(result["is_null"], answer)


def test_rolling_ridge():
    # Test rolling linear regression by comparing it with a manually rolled result.
    # Test on multiple window sizes
    size = 500
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
            pl.Series(name="id", values=list(range(size))),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            - pl.col("x3") * 0.15
            + pds.random() * 0.0001,
        )
    )

    for window_size in [5, 8, 12, 15]:
        df_to_test = df.select(
            "id",
            "y",
            pds.rolling_lin_reg(
                "x1",
                "x2",
                "x3",
                target="y",
                l2_reg=0.1,
                window_size=window_size,
            ).alias("result"),
        ).unnest("result")  # .limit(10)
        df_to_test = df_to_test.filter(pl.col("id") >= window_size - 1).select("coeffs")

        results = []
        for i in range(len(df) - window_size + 1):
            temp = df.slice(i, length=window_size)
            results.append(
                temp.select(pds.lin_reg("x1", "x2", "x3", l2_reg=0.1, target="y").alias("coeffs"))
            )

        df_answer = pl.concat(results)
        assert_frame_equal(df_to_test, df_answer)


def test_hc_lin_reg_report():
    import statsmodels.api as sm

    df = (
        pds.frame(size=1000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
            pds.random(0.0, 1.0).round().cast(pl.UInt8).alias("binary"),
        )
        .with_columns(target=pl.col("x1") * 0.15 + pl.col("x2") * 0.3 + 0.1)
    )

    for se_type in ("se", "hc0", "hc1", "hc2", "hc3"):
        pds_result = df.select(
            pds.lin_reg_report("x1", "x2", "x3", target="target", std_err=se_type).alias("report")
        ).unnest("report")

        Y = df["target"].to_numpy()
        X = df.select("x1", "x2", "x3").to_numpy()

        model = sm.OLS(Y, X)
        results = model.fit()

        if se_type == "se":
            pds_se = pds_result["std_err"].to_numpy()
            sm_se = results.bse
        else:
            pds_se = pds_result[f"{se_type}_se"].to_numpy()
            sm_se = getattr(results, f"{se_type.upper()}_se")

        assert np.all((pds_se - sm_se) < 1e-7)


def test_f32_lin_reg():
    # If they run, they are correct. This is because the underlying functions in src/linalg/ are all
    # generic
    pds.config.LIN_REG_EXPR_F64 = False

    size = 500
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
            pl.Series(name="id", values=list(range(size))),
        )
        .with_columns(
            y=pl.col("x1") * 0.5
            + pl.col("x2") * 0.25
            - pl.col("x3") * 0.15
            + pds.random() * 0.0001,
            y2=pl.col("x1") + pl.col("x2") * 0.3 - pl.col("x3") * 0.1,
        )
    )

    _ = df.select(
        pds.lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
        )
    )

    _ = df.select(
        pds.lin_reg(
            "x1",
            "x2",
            "x3",
            target=["y", "y2"],
        )
    )

    _ = df.select(pds.lin_reg("x1", "x2", "x3", target="y", return_pred=True))

    _ = df.select(pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=0.01, return_pred=False))

    _ = df.select(pds.lin_reg("x1", "x2", "x3", target="y", l2_reg=0.01, return_pred=True))

    _ = df.select(
        pds.lin_reg("x1", "x2", "x3", target="y", l1_reg=0.01, l2_reg=0.01, return_pred=False)
    )

    _ = df.select(pds.lin_reg("x1", "x2", "x3", target=["y", "y2"], l2_reg=0.01, return_pred=False))

    _ = df.select(
        pds.lin_reg(
            "x1",
            "x2",
            "x3",
            target=["y", "y2"],
            l2_reg=0.01,
            null_policy="0.1",  # fill 0.1
        )
    )

    _ = df.select(
        pds.lin_reg_report(
            "x1",
            "x2",
            "x3",
            target="y",
        )
    )

    _ = df.select(pds.lin_reg_report("x1", "x2", "x3", target="y", weights="x1"))

    _ = df.select(
        pds.lin_reg_w_rcond(
            "x1",
            "x2",
            "x3",
            target="y",
            rcond=0.3,
        )
    )

    _ = df.select(
        pds.rolling_lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            window_size=3,
        ).alias("result"),
    ).unnest("result")

    _ = df.select(
        pds.recursive_lin_reg(
            "x1",
            "x2",
            "x3",
            target="y",
            start_with=3,
        ).alias("result"),
    ).unnest("result")

    # Reset it to true
    pds.config.LIN_REG_EXPR_F64 = True


def test_lin_reg_skip_null():
    df = pl.DataFrame(
        {
            "y": [None, 9.5, 10.5, 11.5, 12.5],
            "a": [1, 9, 10, 11, 12],
            "b": [1.0, 0.5, 0.5, 0.5, 0.5],
        }
    )
    res = pl.DataFrame(
        {
            "pred": [float("nan"), 9.5, 10.5, 11.5, 12.5],
            "resid": [float("nan"), 0.0, 0.0, 0.0, 0.0],
        }
    )
    assert_frame_equal(
        df.select(
            pds.lin_reg(
                pl.col("a"), pl.col("b"), target="y", return_pred=True, null_policy="skip"
            ).alias("result")
        ).unnest("result"),
        res,
    )


def test_lin_reg_in_group_by():
    df = pl.DataFrame(
        {
            "A": [1] * 4 + [2] * 4,
            "Y": [1] * 8,
            "X1": [1, 2, 3, 4, 5, 6, 7, 8],
            "X2": [2, 3, 4, 1, 6, 7, 8, 5],
        }
    )

    first = df.filter(pl.col("A").eq(1)).with_columns(
        pds.lin_reg(
            pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
        ).alias("pred")
    )

    second = df.filter(pl.col("A").eq(2)).with_columns(
        pds.lin_reg(
            pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
        ).alias("pred")
    )

    test = (
        df.group_by("A", maintain_order=True)
        .agg(
            "Y",
            "X1",
            "X2",
            pds.lin_reg(
                pl.col("X1"), pl.col("X2"), target=pl.col("Y"), add_bias=False, return_pred=True
            ).alias("pred"),
        )
        .explode("Y", "X1", "X2", "pred")
    )

    test_first = test.filter(pl.col("A") == 1)
    test_second = test.filter(pl.col("A") == 2)

    assert_frame_equal(first, test_first)
    assert_frame_equal(second, test_second)


def test_lin_reg_with_rcond():
    import numpy as np

    size = 5000
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_row_index()
        .with_columns(
            y=pl.col("x1") + pl.col("x2") * 0.2 - 0.3 * pl.col("x3"),
        )
    )

    x = df.select("x1", "x2", "x3").to_numpy()
    y = df.select("y").to_numpy()
    np_coeffs, _, _, np_svs = np.linalg.lstsq(x, y, rcond=0.3)  # default rcond
    np_coeffs = np_coeffs.flatten()

    res = df.select(
        pds.lin_reg_w_rcond(
            "x1",
            "x2",
            "x3",
            target="y",
            rcond=0.3,
        ).alias("result")
    ).unnest("result")
    coeffs = res["coeffs"][0].to_numpy()
    svs = res["singular_values"][0].to_numpy()

    assert np.all(np.abs(coeffs - np_coeffs) < 1e-10)
    assert np.all(np.abs(svs - np_svs) < 1e-10)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": [1, 2, 3, 4, 5], "b": [2, 3, 4, 5, 6]}),
            pl.DataFrame({"j": [2 / 3]}),
        ),
    ],
)
def test_jaccard_col(df, res):
    assert_frame_equal(df.select(pds.jaccard_col("a", "b").alias("j")), res)

    assert_frame_equal(
        df.lazy().select(df.select(pds.jaccard_col("a", "b").alias("j"))).collect(), res
    )


@pytest.mark.parametrize(
    "df, dtype, join_by, res",
    [
        (
            pl.DataFrame(
                {
                    "a": [
                        "0% of my time",
                        "1% to 25% of my time",
                        "75% to 99% of my time",
                        "50% to 74% of my time",
                        "75% to 99% of my time",
                        "50% to 74% of my time",
                    ]
                }
            ),
            pl.Int64,
            "",
            pl.DataFrame({"a": [[0], [1, 25], [75, 99], [50, 74], [75, 99], [50, 74]]}),
        ),
        (
            pl.DataFrame(
                {
                    "a": [
                        "0% of my time",
                        "1% to 25% of my time",
                        "75% to 99% of my time",
                        "50% to 74% of my time",
                        "75% to 99% of my time",
                        "50% to 74% of my time",
                    ]
                }
            ),
            pl.Utf8,
            "-",
            pl.DataFrame({"a": ["0", "1-25", "75-99", "50-74", "75-99", "50-74"]}),
        ),
    ],
)
def test_extract_numbers(df, dtype, join_by, res):
    assert_frame_equal(df.select(pds.extract_numbers("a", join_by=join_by, dtype=dtype)), res)


@pytest.mark.parametrize(
    "df, threshold, res",
    [
        (
            pl.DataFrame({"a": pl.Series([1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0])}),
            0,
            8,
        ),
        (
            pl.DataFrame(
                {
                    "a": pl.Series(
                        [
                            1,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                        ]
                    )
                }
            ),
            0,
            9,
        ),
        (
            pl.DataFrame(
                {
                    "a": pl.Series(
                        [
                            1,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                        ]
                    )
                }
            ),
            0,
            10,
        ),
    ],
)
def test_lempel_ziv_complexity(df, threshold, res):
    test = df.select(pds.query_lempel_ziv(pl.col("a") > threshold, as_ratio=False))
    assert test.item(0, 0) == res


def test_ks_stats():
    from scipy.stats import ks_2samp
    import numpy as np

    a = np.random.random(size=1000)
    b = np.random.random(size=1000)
    df = pl.DataFrame({"a": a, "b": b})

    stats = ks_2samp(a, b).statistic
    # Only statistic for now
    res = df.select(pds.ks_2samp("a", "b").struct.field("statistic")).item(0, 0)

    assert np.isclose(stats, res)


@pytest.mark.parametrize(
    "df, k, dist, res",
    [
        (
            pl.DataFrame(dict(x=[1, 2, 10], y=[2, 5, 10])),
            1,
            "l2",
            5.675998737756144,
        ),
    ],
)
def test_knn_entropy(df, k, dist, res):
    test = df.select(pds.query_knn_entropy("x", "y", k=k, dist=dist)).item(0, 0)

    assert np.isclose(test, res)

    test_par = df.select(pds.query_knn_entropy("x", "y", k=k, dist=dist, parallel=True)).item(0, 0)

    assert np.isclose(test_par, res)


@pytest.mark.parametrize(
    "df, eq_var",
    [
        (
            pl.DataFrame({"a": np.random.normal(size=1_000), "b": np.random.normal(size=1_000)}),
            True,
        ),
        (
            pl.DataFrame({"a": np.random.normal(size=1_000), "b": np.random.normal(size=1_000)}),
            False,
        ),
    ],
)
def test_ttest_ind(df, eq_var):
    from scipy.stats import ttest_ind

    res = df.select(pds.ttest_ind("a", "b", equal_var=eq_var))
    res = res.item(0, 0)  # A dictionary
    statistic = res["statistic"]
    pvalue = res["pvalue"]

    scipy_res = ttest_ind(df["a"].to_numpy(), df["b"].to_numpy(), equal_var=eq_var)

    assert np.isclose(statistic, scipy_res.statistic)
    assert np.isclose(pvalue, scipy_res.pvalue)


@pytest.mark.parametrize(
    "df",
    [
        (
            pl.DataFrame(
                {
                    "a": np.random.normal(size=1_000),
                    "b": list(np.random.normal(size=998)) + [None, None],
                }
            )
        ),
    ],
)
def test_welch_t(df):
    from scipy.stats import ttest_ind

    res = df.select(pds.ttest_ind("a", "b", equal_var=False))
    res = res.item(0, 0)  # A dictionary
    statistic = res["statistic"]
    pvalue = res["pvalue"]

    s1 = df["a"].drop_nulls().to_numpy()
    s2 = df["b"].drop_nulls().to_numpy()
    scipy_res = ttest_ind(s1, s2, equal_var=False)

    assert np.isclose(statistic, scipy_res.statistic)
    assert np.isclose(pvalue, scipy_res.pvalue)


@pytest.mark.parametrize(
    "df",
    [
        (
            pl.DataFrame(
                {
                    "x": ["a"] * 200 + ["b"] * 300 + ["c"] * 500 + ["d"] * 200,
                    "y": [1] * 800 + [2] * 400,
                }
            )
        ),
        (
            pl.DataFrame(
                {
                    "x": np.random.RandomState(42).permutation([0] * 2**16 + [1] * 2**16),
                    "y": np.random.RandomState(43).permutation([0] * 2**16 + [1] * 2**16),
                }
            )
        ),
    ],
)
def test_chi2(df: pl.DataFrame):
    import pandas as pd
    from scipy.stats import chi2_contingency

    res = df.select(pds.chi2("x", "y")).item(0, 0)
    stats, p = res["statistic"], res["pvalue"]

    df2 = df.to_pandas()
    contigency = pd.crosstab(index=df2["x"], columns=df2["y"])
    sp_res = chi2_contingency(contigency.to_numpy(), correction=False)
    sp_stats, sp_p = sp_res.statistic, sp_res.pvalue
    assert np.isclose(stats, sp_stats)
    assert np.isclose(p, sp_p)


@pytest.mark.parametrize(
    "df",
    [
        (pl.DataFrame({"a": np.random.random(size=100)})),
    ],
)
def test_normal_test(df):
    from scipy.stats import normaltest

    res = df.select(pds.normal_test("a"))
    res = res.item(0, 0)  # A dictionary
    statistic = res["statistic"]
    pvalue = res["pvalue"]

    scipy_res = normaltest(df["a"].to_numpy())

    assert np.isclose(statistic, scipy_res.statistic)
    assert np.isclose(pvalue, scipy_res.pvalue)


@pytest.mark.parametrize(
    "df",
    [
        (pl.DataFrame({"a": 1000 * np.random.random(size=100)})),
    ],
)
def test_expit(df):
    from scipy.special import expit

    res = df.select(pds.expit("a"))["a"].to_numpy()
    scipy_res = expit(df["a"].to_numpy())
    assert np.isclose(res, scipy_res, equal_nan=True).all()


@pytest.mark.parametrize(
    "df",
    [
        (pl.DataFrame({"a": [0.0, 1.0, 2.0] + list(np.random.random(size=100))})),
    ],
)
def test_logit(df):
    from scipy.special import logit

    res = df.select(pds.logit("a"))["a"].to_numpy()
    scipy_res = logit(df["a"].to_numpy())
    assert np.isclose(res, scipy_res, equal_nan=True).all()


@pytest.mark.parametrize(
    "df",
    [
        (pl.DataFrame({"a": [0.0] + list(100 * np.random.random(size=100))})),
    ],
)
def test_gamma(df):
    from scipy.special import gamma

    res = df.select(pds.gamma("a"))["a"].to_numpy()
    scipy_res = gamma(df["a"].to_numpy())
    assert np.isclose(res, scipy_res, equal_nan=True).all()


@pytest.mark.parametrize(
    "df, dist, k, res",
    [
        (
            pl.DataFrame({"id": range(5), "val1": range(5), "val2": range(5), "val3": range(5)}),
            "l2",
            2,
            # Remember that this counts self as well.
            pl.DataFrame({"nn": [[0, 1, 2], [1, 2, 0], [2, 1, 3], [3, 2, 4], [4, 3, 2]]}),
        ),
    ],
)
def test_knn_ptwise(df, dist, k, res):
    df2 = df.select(
        pds.query_knn_ptwise(
            pl.col("val1"), pl.col("val2"), pl.col("val3"), index="id", dist=dist, k=k
        )
        .list.eval(
            pl.element().sort().cast(pl.UInt32)
        )  # sort to prevent false results on equidistant neighbors
        .alias("nn")
    )
    res = res.select(pl.col("nn").list.eval(pl.element().sort().cast(pl.UInt32)))
    assert_frame_equal(df2, res)


@pytest.mark.parametrize(
    "df, dist, k, res",
    [
        (
            pl.DataFrame(
                {
                    "id": [0, 1, 2, 3, 4, 5],
                    "values": [0, 1, 2, 3, 4, 5],
                    "a": [0.1, 1, 10, 100, float("nan"), 1.0],
                    "b": [0.15, 1.5, 15, 150, 1.0, None],
                    "c": [0.12, 1.2, 12, 120, 2.0, 2.0],
                }
            ),
            "sql2",
            2,
            pl.Series(name="knn_avg", values=[1.5, 1.0, 0.5, 1.5, None, None]),
        ),
    ],
)
def test_knn_avg(df, dist, k, res):
    to_test = df.select(
        pds.query_knn_avg(
            "a",
            "b",
            "c",
            target="values",  # will be casted to f64
            k=2,
            dist="sql2",
            weighted=False,
        ).alias("knn_avg")
    )

    assert_series_equal(to_test["knn_avg"], res)


@pytest.mark.parametrize(
    "df, dist, k, max_bound, res",
    [
        (
            pl.DataFrame(
                {
                    "id": [0, 1, 2, 3],
                    "a": [0.1, 1, 10, 100],
                    "b": [0.15, 1.5, 15, 150],
                    "c": [0.12, 1.2, 12, 120],
                }
            ),
            "sql2",
            2,
            4.0,
            pl.DataFrame({"friends": [[0, 1], [1, 0], [2], [3]]}),
        ),
    ],
)
def test_knn_ptwise_max_bound(df, dist, k, max_bound, res):
    df2 = df.select(
        pds.query_knn_ptwise(
            "a",
            "b",
            "c",
            index="id",
            k=k,
            dist="sql2",
            max_bound=max_bound,
        ).alias("friends")
    )
    res = res.select(pl.col("friends").list.eval(pl.element().cast(pl.UInt32)))
    assert_frame_equal(df2, res)


def test_knn_ptwise_skip():
    df = pl.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "a1": [0.1, 0.2, 0.3, 0.4],
            "a2": [0.1, 0.2, 0.3, 0.4],
            "a3": [0.1, 0.2, 0.3, 0.4],
            "can_eval": [1, 0, 1, 1],
        }
    )

    res1 = df.with_columns(
        pds.query_knn_ptwise(
            "a1",
            "a2",
            "a3",
            index="id",
            k=1,
            dist="l2",  # squared l2
            parallel=False,
            eval_mask=pl.col("can_eval") == 1,
        ).alias("best friends")
    )
    friends = list(res1["best friends"])
    assert friends[1] is None  # should be true because we are skipping this evaluation
    assert list(friends[0]) == [0, 1]  # id 1 can still be a neighbor

    res2 = df.with_columns(
        pds.query_knn_ptwise(
            "a1",
            "a2",
            "a3",
            index="id",
            k=1,
            dist="l2",  # squared l2
            parallel=False,
            data_mask=pl.col("can_eval") == 1,
        ).alias("best friends")
    )
    friends = list(res2["best friends"])
    assert list(friends[0]) == [0, 2]  # id 1 cannot be a neighbor anymore
    all_neighbors = []
    for f in friends:
        all_neighbors += list(f)
    assert 1 not in set(all_neighbors)  # id 1 cannot be a neighbor anymore

    res3 = df.with_columns(
        pds.query_knn_ptwise(
            "a1",
            "a2",
            "a3",
            index="id",
            k=1,
            dist="l2",  # squared l2
            parallel=False,
            eval_mask=pl.col("can_eval") == 1,
            data_mask=pl.col("can_eval") == 1,
        ).alias("best friends")
    )
    friends = list(res3["best friends"])
    # Now 1 cannot be a neighbor, and we don't evaluate for 1
    assert friends[1] is None  # should be true because we are skipping this evaluation
    assert list(friends[0]) == [0, 2]  # id 1 can still be a neighbor


@pytest.mark.parametrize(
    "df, x, dist, k, res",
    [
        (  # Only the second is the nearest neighbor (l2 sense) to [0.5, 0.5, 0.5]
            pl.DataFrame(
                {
                    "id": [1, 2, 3],
                    "val1": [0.1, 0.2, 5.0],
                    "val2": [0.1, 0.3, 10.0],
                    "val3": [0.1, 0.4, 11.0],
                }
            ),
            [0.5, 0.5, 0.5],
            "l2",
            1,
            pl.DataFrame({"id": [2]}),
        ),
        (  # If cosine dist, the first would be the nearest
            pl.DataFrame(
                {
                    "id": [1, 2, 3],
                    "val1": [0.1, 0.2, 5.0],
                    "val2": [0.1, 0.3, 10.0],
                    "val3": [0.1, 0.4, 11.0],
                }
            ),
            [0.5, 0.5, 0.5],
            "cosine",
            1,
            pl.DataFrame({"id": [1]}),
        ),
    ],
)
def test_is_knn_from(df, x, dist, k, res):
    test = df.filter(
        pds.is_knn_from(pl.col("val1"), pl.col("val2"), pl.col("val3"), pt=x, dist=dist, k=k)
    ).select(pl.col("id"))

    assert_frame_equal(test, res)


@pytest.mark.parametrize(
    "df, dist, res",
    [
        (
            pl.DataFrame(
                {
                    "id": [1, 2, 3],
                    "val1": [0.1, 0.2, 5.0],
                    "val2": [0.1, 0.3, 10.0],
                    "val3": [0.1, 0.4, 11.0],
                }
            ),
            "sql2",
            pl.DataFrame({"id": [[1, 2], [2, 1], [3]]}),
        ),
    ],
)
def test_radius_ptwise(df, dist, res):
    test = df.select(
        pds.query_radius_ptwise("val1", "val2", "val3", dist=dist, r=0.3, index="id").alias("id")
    ).explode("id")  # compare after explode
    res = res.explode("id").select(pl.col("id").cast(pl.UInt32))
    assert_frame_equal(test, res)


@pytest.mark.parametrize(
    "df, r, dist, res",
    [
        (
            pl.DataFrame({"x": range(5), "y": range(5), "z": range(5)}),
            4,
            "sql2",
            pl.DataFrame({"nb_cnt": [2, 3, 3, 3, 2]}),  # A point is always its own neighbor
        ),
        (
            pl.DataFrame(
                {
                    "x": [0.1, 0.2, 0.5, 0.9, 2.1],
                    "y": [0.1, 0.3, 0.6, 1.1, 3.3],
                    "z": [0.1, 0.4, 0.8, 1.2, 4.1],
                }
            ),
            1.0,
            "l1",
            pl.DataFrame({"nb_cnt": [2, 3, 2, 1, 1]}),
        ),
    ],
)
def test_nb_cnt(df, r, dist, res):
    test = df.select(
        pds.query_nb_cnt(
            pl.col("x"),
            pl.col("y"),
            pl.col("z"),  # Columns used as the coordinates in n-d space
            r=r,
            dist=dist,
        )
        .cast(pl.UInt32)
        .alias("nb_cnt")
    )
    assert_frame_equal(test, res.with_columns(pl.col("nb_cnt").cast(pl.UInt32)))


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame(
                {
                    "x1": [51.5007],
                    "x2": [0.1246],
                    "y1": [40.6892],
                    "y2": [74.0445],
                }
            ),
            pl.DataFrame({"dist": [5574.840456848555]}),
        ),
    ],
)
def test_haversine(df, res):
    test = df.select(
        pds.haversine(pl.col("x1"), pl.col("x2"), pl.col("y1"), pl.col("y2")).alias("dist")
    )
    assert_frame_equal(test, res)


@pytest.mark.parametrize(
    "s, res",
    [
        (list(range(100)), 0.010471299867295437),
        (np.sin(2 * np.pi * np.arange(3000) / 100), 0.16367903754688098),
    ],
)
def test_sample_entropy(s, res):
    # Test 1's answer comes from comparing result with Tsfresh
    # Thest 2's answer comes from running this using the Python code on Wikipedia
    df = pl.Series(name="a", values=s).to_frame()
    entropy = df.select(pds.query_sample_entropy(pl.col("a"))).item(0, 0)
    assert np.isclose(entropy, res, atol=1e-12, equal_nan=True)


@pytest.mark.parametrize(
    "s, m, r, scale, res",
    [
        ([12, 13, 15, 16, 17] * 10, 2, 0.9, True, 0.282456191276673),
        ([1.4, -1.3, 1.7, -1.2], 2, 0.5, False, 0.0566330122651324),
        (
            [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1],
            2,
            0.5,
            False,
            0.002223871246127107,
        ),
        (
            [0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1, 1, 0, 0, 1],
            2,
            0.5,
            False,
            0.47133806162842484,
        ),
        ([85, 80, 89] * 17, 2, 3, False, 1.099654110658932e-05),
        ([85, 80, 89] * 17, 2, 3, True, 0.0),
    ],
)
def test_approximate_entropy(s, m, r, scale, res):
    df = pl.Series(name="a", values=s).to_frame()

    entropy = df.select(
        pds.query_approx_entropy("a", m=m, filtering_level=r, scale_by_std=scale)
    ).item(0, 0)
    assert np.isclose(entropy, res, atol=1e-12, equal_nan=True)


def test_approximate_entropy_edge_cases():
    df = pl.Series(name="a", values=[1]).to_frame()
    res = df.select(pds.query_approx_entropy("a", m=2, filtering_level=0.1, scale_by_std=False))
    assert np.isnan(res.item(0, 0))


@pytest.mark.parametrize(
    "df, n_bins, res",
    [
        (  # 4 * (0.1 - 0.0001) * np.log(0.1 / 0.0001) + (0.1 - 0.5) * np.log(0.1 / 0.5)
            pl.DataFrame({"ref": range(1000), "act": list(range(500)) + [600] * 500}),
            10,
            3.4041141744549024,
        ),
    ],
)
def test_psi(df, n_bins, res):
    ans = df.select(pds.psi("act", pl.col("ref"), n_bins=n_bins)).item(0, 0)
    assert np.isclose(ans, res)


@pytest.mark.parametrize(
    "df, res",
    [
        (  # discrete version of the above example
            pl.DataFrame(
                {
                    "ref": [0] * 100
                    + [1] * 100
                    + [2] * 100
                    + [3] * 100
                    + [4] * 100
                    + [5] * 100
                    + [6] * 100
                    + [7] * 100
                    + [8] * 100
                    + [9] * 100,
                    "act": [0] * 100 + [1] * 100 + [2] * 100 + [3] * 100 + [4] * 100 + [6] * 500,
                }
            ),
            3.4041141744549024,
        ),
    ],
)
def test_psi_discrete(df, res):
    ans = df.select(pds.psi_discrete("act", pl.col("ref"))).item(0, 0)
    assert np.isclose(ans, res)


def test_query_similar_count():
    def square_l2_distance(a: np.ndarray, b: np.ndarray):
        diff = a - b
        return diff.dot(diff)

    size = 1000
    df = pds.frame(size=size).select(
        pds.random(0.0, 1.0).alias("x1"),
    )

    query = np.random.random(size=3)

    cnt = df.select(
        pds.query_similar_count(query=query, target="x1", metric="sqzl2", threshold=0.1)
    ).item(0, 0)

    x1 = df["x1"].to_numpy()
    actual_cnt = 0
    z_normed_query = (query - np.mean(query)) / np.std(query, ddof=1)
    for i in range(0, len(x1) - len(query) + 1):
        sl = x1[i : i + len(query)]
        znormed = (sl - np.mean(sl)) / np.std(sl, ddof=1)
        actual_cnt += int((square_l2_distance(znormed, z_normed_query) < 0.1))

    assert cnt == actual_cnt

    cnt = df.select(
        pds.query_similar_count(query=query, target="x1", metric="sql2", threshold=0.1)
    ).item(0, 0)

    x1 = df["x1"].to_numpy()
    actual_cnt = 0
    for i in range(0, len(x1) - len(query) + 1):
        sl = x1[i : i + len(query)]
        actual_cnt += int((square_l2_distance(sl, query) < 0.1))

    assert cnt == actual_cnt


def test_auto_corr():
    def autocorr(x, lag):
        mean = np.mean(x)
        v = np.var(x)
        xp = x - mean
        return np.sum(xp[lag:] * xp[:-lag]) / ((len(x) - lag) * v)

    df = pds.frame(size=2_000).select(
        pds.random(0.0, 1.0).alias("x1"),
    )
    x = df["x1"].to_numpy()

    pds_result = df.select(
        *(pds.query_auto_corr("x1", lag=i).alias(str(i)) for i in range(1, 10))
    ).row(0)
    pds_result = np.array(pds_result)
    np_result = np.array([autocorr(x, i) for i in range(1, 10)])

    assert np.all(np.abs(pds_result - np_result) < 1e-6)


@pytest.mark.parametrize(
    "df",
    [
        (
            pl.DataFrame(
                {
                    "a": [1.5, 1.0, 4.0, 6.0, 5.7, 5.0, 7.8, 9.0, 7.5, 9.5, 9.0],
                    "b": [1, 1, 2, 2, 1, 1, 3, 3, 1, 1, 1],
                }
            )
        ),
    ],
)
def test_isotonic(df):
    from scipy.optimize import isotonic_regression

    pds_no_weights = df.select(res=pds.isotonic_regression(pl.col("a"), weights=None))[
        "res"
    ].to_numpy()

    scipy_no_weights = isotonic_regression(df["a"].to_numpy()).x

    assert np.all(pds_no_weights == scipy_no_weights)

    pds_w_weights = df.select(res=pds.isotonic_regression(pl.col("a"), weights=pl.col("b")))[
        "res"
    ].to_numpy()

    scipy_w_weights = isotonic_regression(df["a"].to_numpy(), weights=df["b"].to_numpy()).x

    assert np.all(pds_w_weights == scipy_w_weights)


def test_next_up_down():
    a = np.random.random(size=100)
    df = pl.DataFrame({"a": a})
    a_up = df.select(pds.next_up("a"))["a"].to_numpy()
    a_down = df.select(pds.next_down("a"))["a"].to_numpy()

    assert np.all(np.nextafter(a, 1.0) == a_up)
    assert np.all(np.nextafter(a, 0.0) == a_down)


def test_xlogy():
    df = pl.DataFrame(
        {
            "a": [0.0, 0.0, float("nan"), 3.0],
            "b": [1.0, float("nan"), 1.0, 4.0],
        }
    )
    # a = 0 and b is not nan, then a * log(b) = 0
    # otherwise, do a * log(b)
    answer = [0.0, float("nan"), float("nan"), float(3.0 * np.log(4.0))]
    pds_result = df.select(res=pds.xlogy("a", "b"))["res"].to_numpy()
    assert np.allclose(pds_result, answer, equal_nan=True)


def test_digamma():
    import scipy

    a = np.random.random(size=100)
    df = pl.DataFrame({"a": a})
    pds_digamma = df.select(pds.digamma(pl.col("a")))["a"].to_numpy()
    scipy_digamma = scipy.special.psi(a)

    assert np.all(np.isclose(pds_digamma, scipy_digamma, atol=1e-5))


def test_kth_nb_dist():
    size = 2000
    df = pl.DataFrame(
        {
            "id": range(size),
        }
    ).with_columns(
        pds.random().alias("var1"),
        pds.random().alias("var2"),
        pds.random().alias("var3"),
    )
    # method 1 is what we want to test
    # method 2 is assumed to be the truth.
    test = (
        df.select(
            pds.query_dist_from_kth_nb("var1", "var2", "var3", dist="l1", k=3).alias(
                "kth_nb_dist_method_1"
            ),
            pds.query_knn_ptwise(
                "var1", "var2", "var3", index="id", return_dist=True, k=3, dist="l1"
            )
            .struct.field("dist")
            .list.last()
            .alias("kth_nb_dist_method_2"),
        )
        .select((pl.col("kth_nb_dist_method_1") == pl.col("kth_nb_dist_method_2")).all())
        .item(0, 0)
    )

    assert test is True


#


def test_combinations():
    df = pl.DataFrame({"category": ["a", "a", "a", "b", "b"], "values": [1, 2, 3, 4, 5]})

    result = df.select(pds.combinations("category", 2, unique=True))

    answer = pl.DataFrame({"category": [["a", "b"]]})

    assert_frame_equal(result, answer)

    try:
        _ = df.select(pds.combinations("category", 3, unique=True))
        assert False  # Should not reach here. 2 uniques, 2 choose 3 is not possible.
    except:  # noqa : E722
        assert True

    result = df.group_by("category").agg(pds.combinations("values", 2)).sort("category")

    answer = pl.DataFrame({"category": ["a", "b"], "values": [[[1, 2], [1, 3], [2, 3]], [[4, 5]]]})

    assert_frame_equal(result, answer)


@pytest.mark.parametrize(
    "df, ans",
    [
        (
            pl.DataFrame({"a": [1, None, None], "b": [1, 2, 3]}),
            pl.DataFrame({"product": [[1, 1], [1, 2], [1, 3]]}),
        ),
    ],
)
def test_product(df, ans):
    result = df.select(pds.product("a", "b").alias("product"))
    assert_frame_equal(result, ans)


def test_weighted_corr():
    import numpy as np

    def weighted_mean(x, w):
        """Calculates the weighted mean of an array."""
        return np.sum(x * w) / np.sum(w)

    def weighted_covariance(x, y, w):
        """Calculates the weighted covariance between two arrays."""
        mean_x = weighted_mean(x, w)
        mean_y = weighted_mean(y, w)
        return np.sum(w * (x - mean_x) * (y - mean_y)) / np.sum(w)

    def weighted_correlation(x, y, w):
        """Calculates the weighted Pearson correlation coefficient."""
        cov_xy = weighted_covariance(x, y, w)
        cov_xx = weighted_covariance(x, x, w)
        cov_yy = weighted_covariance(y, y, w)

        # Handle cases where variance is zero to avoid division by zero
        if cov_xx == 0 or cov_yy == 0:
            return 0.0  # Or handle as appropriate for your application

        return cov_xy / np.sqrt(cov_xx * cov_yy)

    df = pds.frame().select(
        pds.random().alias("a1"), pds.random().alias("a2"), pds.random().alias("w")
    )

    a1 = df["a1"].to_numpy()
    a2 = df["a2"].to_numpy()
    w = df["w"].to_numpy()

    np_result = weighted_correlation(a1, a2, w)
    pds_result = df.select(pds.weighted_corr("a1", "a2", "w")).item(0, 0)

    np.isclose(np_result, pds_result, atol=1e-8)



import polars as pl
import polars_ds as pds
import polars_ds.pipeline.transforms as t
import pytest
from polars.testing import assert_frame_equal


def test_ordinal_encoding():
    from polars_ds.pipeline import Blueprint

    df_fit = pl.DataFrame({"gender": ["M", "F", "F", None], "data": [1, 2, 3, 4]})

    df_non_fit = pl.DataFrame({"gender": ["M", "X", "G", "F", None], "data": [1, 2, 3, 4, 5]})

    bp1 = Blueprint(df_fit, name="example", lowercase=True).ordinal_encode(
        cols=["gender"], null_value=-1, unknown_value=-2
    )

    pipe1 = bp1.materialize()
    result1_fit = pipe1.transform(df_fit)["gender"].to_list()
    # M = 1, F = 0, null = -1, unknown = -2
    assert result1_fit == [1.0, 0.0, 0.0, -1.0]
    result1_non_fit = pipe1.transform(df_non_fit)["gender"].to_list()
    assert result1_non_fit == [1.0, -2.0, -2.0, 0.0, -1.0]

    bp2 = Blueprint(df_fit, name="example", lowercase=True).ordinal_encode(cols=["gender"])
    # by default, nulls remain null
    pipe2 = bp2.materialize()
    result2 = pipe2.transform(df_fit)["gender"].to_list()
    assert result2 == [1.0, 0.0, 0.0, None]


def test_linear_impute():
    df = pl.DataFrame(
        {
            "a": [3, 2, 3, 4, 5, 6, 7, 8, 9, 11],
            "b": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "c": [4.0, 4.0, None, 8.0, 10.0, 12.0, 14.0, 16.0, 18.0, 21.0],
        }
    )

    imputed_c = df.with_columns(
        c=t.linear_impute(df, features=["a", "b"], target="c")[0].cast(pl.Float64)
    ).select("c")

    correct_c = df.select(c=(pl.col("a") + pl.col("b")).cast(pl.Float64))

    assert_frame_equal(imputed_c, correct_c)


def test_conditional_impute():
    df = pl.DataFrame(
        {
            "a": [float("nan"), None, float("inf"), 9999, 100, 100, 100, 800],
        }
    )

    res = df.with_columns(
        t.conditional_impute(
            df,
            {"a": ((pl.col("a").is_finite().not_()) | pl.col("a").is_null() | (pl.col("a") > 899))},
            method="mean",
        )[0].alias("result")
    )["result"]

    assert list(res)[:4] == [275.0, 275.0, 275.0, 275.0]

    res = df.with_columns(
        t.conditional_impute(
            df,
            {"a": ((pl.col("a").is_finite().not_()) | pl.col("a").is_null() | (pl.col("a") > 899))},
            method="median",
        )[0].alias("result")
    )["result"]

    assert list(res)[:4] == [100.0, 100.0, 100.0, 100.0]


def test_winsorize():
    df = pds.frame(size=1000).select(
        pds.random(0.0, 1.0).alias("x1"),
    )

    q_low = 0.05
    q_high = 0.95

    should_be_true = (
        df.with_columns(x2=t.winsorize(df, ["x1"], q_low=q_low, q_high=q_high)[0])
        .select(
            # the max and min of x2 should be 0.95, 0.05 - percentile of x1
            x1=pl.col("x2").max() == pl.col("x1").quantile(q_high),
            x2=pl.col("x2").min() == pl.col("x1").quantile(q_low),
        )
        .row(0)
    )

    assert all(should_be_true)


def test_polynomial_features_deg2():
    df = pds.frame(size=1000).select(
        pds.random(0.0, 1.0).alias("x1"),
        pds.random(0.0, 1.0).alias("x2"),
    )

    df_test = df.select(
        result=t.polynomial_features(["x1", "x2"], degree=2, interaction_only=True)[0]
    )

    df_correct = df.select(result=pl.col("x1") * pl.col("x2"))

    assert_frame_equal(df_test, df_correct)

    df_test = df.select(t.polynomial_features(["x1", "x2"], degree=2))

    df_correct = df.select(
        (pl.col("x1") * pl.col("x1")).alias("x1*x1"),
        (pl.col("x1") * pl.col("x2")).alias("x1*x2"),
        (pl.col("x2") * pl.col("x2")).alias("x2*x2"),
    )

    assert_frame_equal(df_test, df_correct)


def test_robust_scale():
    q_low = 0.25
    q_high = 0.75

    df = pds.frame(size=1000).select(
        pds.random(0.0, 1.0).alias("x1"),
    )

    method = "midpoint"

    robust_x1 = df.select(
        x2=t.robust_scale(df, ["x1"], q_low=q_low, q_high=q_high, method=method)[0]
    )

    robust_correct = df.select(
        x2=(pl.col("x1") - pl.col("x1").quantile(q_low, interpolation=method))
        / (
            pl.col("x1").quantile(q_high, interpolation=method)
            - pl.col("x1").quantile(q_low, interpolation=method)
        )
    )

    assert_frame_equal(robust_x1, robust_correct)


def test_one_hot_encode_options():
    df = pl.DataFrame(
        {
            "a": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "b": [None, "a", "a", "c", "d", "a", "a", None, "b", "c"],
        }
    )

    # polars_ds's one hot encode transform ignores nulls all the time
    # if you want nulls to be one-hot-encoded, use col('abc').is_null()

    df_t1 = df.with_columns(t.one_hot_encode(df, cols=["b"]))

    columns = df_t1.columns
    assert columns == ["a", "b", "b_a", "b_b", "b_c", "b_d"]  # the new columns are in lex order

    df_t2 = df.with_columns(t.one_hot_encode(df, cols=["b"], separator="|"))
    columns = df_t2.columns
    assert columns == ["a", "b", "b|a", "b|b", "b|c", "b|d"]  # the new columns are in lex order

    df_t3 = df.with_columns(t.one_hot_encode(df, cols=["b"], separator="|", drop_first=True))
    columns = df_t3.columns
    assert columns == ["a", "b", "b|b", "b|c", "b|d"]  # Drops the first


def test_target_encode():
    from category_encoders import TargetEncoder

    # random df to test target_encode
    df = pds.frame(size=2000).select(
        pds.random_int(0, 3).cast(pl.String).alias("cat"), pds.random_int(0, 2).alias("target")
    )

    # polars_ds target encode
    df1 = df.select(
        t.target_encode(df, ["cat"], target="target", min_samples_leaf=20, smoothing=10.0)
    )

    # category_encoder target encode
    df_pd = df.to_pandas()
    y = df_pd["target"]
    cols = ["cat"]
    enc = TargetEncoder(min_samples_leaf=20, smoothing=10.0, cols=cols)
    enc.fit(df_pd[cols], y)
    df_transformed = enc.transform(df_pd[cols])
    df2 = pl.from_pandas(df_transformed[cols])  # output from category_encoders

    assert_frame_equal(df1, df2)

    # Test again with different min_sample_leaf and smoothing

    # random df to test target_encode
    df = pds.frame(size=2000).select(
        pds.random_int(0, 3).cast(pl.String).alias("cat"), pds.random_int(0, 2).alias("target")
    )

    # polars_ds target encode
    df1 = df.select(
        t.target_encode(df, ["cat"], target="target", min_samples_leaf=30, smoothing=5.0)
    )

    # category_encoder target encode
    df_pd = df.to_pandas()
    y = df_pd["target"]
    cols = ["cat"]
    enc = TargetEncoder(min_samples_leaf=30, smoothing=5.0, cols=cols)
    enc.fit(df_pd[cols], y)
    df_transformed = enc.transform(df_pd[cols])
    df2 = pl.from_pandas(df_transformed[cols])  # output from category_encoders
    assert_frame_equal(df1, df2)


def test_woe_encode():
    from category_encoders import WOEEncoder

    df = pds.frame(size=2000).select(
        pds.random_int(0, 3).cast(pl.String).alias("str_col"),
        pds.random_int(0, 2).alias("target"),
    )
    df1 = df.select(t.woe_encode(df, ["str_col"], target="target"))

    X = df["str_col"].to_numpy()
    y = df["target"].to_numpy()

    enc = WOEEncoder()
    df2 = pl.from_pandas(enc.fit_transform(X, y))
    df2.columns = ["str_col"]

    assert_frame_equal(df1, df2)


def test_scale():
    from sklearn.preprocessing import StandardScaler, MinMaxScaler

    df = pds.frame(size=100).select(
        pds.random().alias("a"),
        pds.random_normal(0.0, 1.0).alias("b"),
    )

    cols = df.columns
    df1 = df.with_columns(t.scale(df, cols=cols, method="standard"))

    scaler = StandardScaler()
    mat = scaler.fit_transform(df.to_pandas())
    df2 = pl.DataFrame(mat)
    df2.columns = cols

    assert_frame_equal(df1, df2)

    df1 = df.with_columns(t.scale(df, cols=cols, method="min_max"))

    scaler = MinMaxScaler()
    mat = scaler.fit_transform(df.to_pandas())
    df2 = pl.DataFrame(mat)
    df2.columns = cols

    assert_frame_equal(df1, df2)


@pytest.mark.parametrize(
    "df",
    [
        (pl.DataFrame({"a": [1, None, 2, 3, 3, 3, 3, 4], "b": [3, None, None, 3, 2, 2, 1, 4]})),
    ],
)
def test_impute(df):
    # all columns will be tested

    from sklearn.impute import SimpleImputer

    cols = df.columns
    df1 = df.with_columns(t.impute(df, cols=cols, method="mean"))

    imputer = SimpleImputer(strategy="mean")
    mat = imputer.fit_transform(df.to_pandas())
    df2 = pl.from_numpy(mat, schema=cols)

    assert_frame_equal(df1, df2)

    df1 = df.with_columns(t.impute(df, cols=cols, method="median"))

    imputer = SimpleImputer(strategy="median")
    mat = imputer.fit_transform(df.to_pandas())
    df2 = pl.from_numpy(mat, schema=cols)

    assert_frame_equal(df1, df2)


@pytest.mark.parametrize(
    "df, ranking",
    [
        (
            pl.DataFrame(
                {
                    "col": [
                        "bad",
                        "bad",
                        "good",
                        "neutral",
                        "neutral",
                        "neutral",
                        "bad",
                        "good",
                        None,
                        "unknown",
                    ]
                }
            ),
            ["bad", "neutral", "good"],
        ),
    ],
)
def test_rank_hot_encode(df, ranking):
    # all columns will be tested against the native Python lambda impl
    # Naive Python Lambda method
    mapping = dict(zip(ranking, range(len(ranking))))
    df_naive = df.select(
        (
            (
                pl.col("col").map_elements(
                    lambda x: mapping[x] if x in mapping else None,
                    return_dtype=pl.Int32,
                )
                >= i
            )
            .cast(pl.Int8)
            .alias(f"col>={ranking[i]}")
            # fill_null with -1. If we get null, that means the value is not in ranking or is null.
        )
        for i in range(1, len(ranking))
    )

    df_to_test = df.select(t.rank_hot_encode(col="col", ranking=ranking))
    assert_frame_equal(df_to_test, df_naive)



"""Linear Regression Related Expressions in Polars."""

from __future__ import annotations
import polars as pl
import warnings
from typing import List, Any, Literal

# Internal dependencies
from polars_ds.typing import LRSolverMethods, NullPolicy
from polars_ds._utils import pl_plugin
import polars_ds.config as cfg

__all__ = [
    "lin_reg",
    "lin_reg_w_rcond",
    "simple_lin_reg",
    "recursive_lin_reg",
    "rolling_lin_reg",
    "lin_reg_report",
]


# Despite the typing requirments in the function signatures, we allow some slack
# by accepting the most common Series/Array types.
def lr_formula(s: Any) -> pl.Expr:
    if isinstance(s, str):
        return pl.sql_expr(s).alias(s)
    elif isinstance(s, pl.Series):
        return pl.lit(s)
    elif isinstance(s, pl.Expr):
        return s
    elif hasattr(s, "__array__"):
        return pl.lit(pl.Series(values=s.__array__()))
    else:
        raise ValueError(
            "Input can only be str or polars expression. The str must be valid SQL strings that polars can understand."
        )


def simple_lin_reg(
    x: str | pl.Expr,
    target: str | pl.Expr,
    add_bias: bool = False,
    weights: str | pl.Expr | None = None,
    return_pred: bool = False,
) -> pl.Expr:
    """
    Simple least square with 1 predictive variable and 1 target.

    Parameters
    ----------
    x
        The variables used to predict target
    target
        The target variable
    add_bias
        Whether to add a bias term
    weights
        Whether to perform a weighted linear regression or not.
    return_pred
        If true, return prediction and residue. If false, return coefficients. Note that
        for coefficients, it reduces to one output (like max/min), but for predictions and
        residue, it will return the same number of rows as in input.
    """
    # No test. All forumla here are mathematically correct.
    xx = lr_formula(x)
    yy = lr_formula(target)
    if add_bias:
        if weights is None:
            x_mean = xx.mean()
            y_mean = yy.mean()
            beta = (xx - x_mean).dot(yy - y_mean) / (xx - x_mean).dot(xx - x_mean)
            alpha = y_mean - beta * x_mean
        else:
            w = lr_formula(weights)
            w_sum = w.sum()
            x_wmean = w.dot(xx) / w_sum
            y_wmean = w.dot(yy) / w_sum
            beta = w.dot((xx - x_wmean) * (yy - y_wmean)) / (w.dot((xx - x_wmean).pow(2)))
            alpha = y_wmean - beta * x_wmean

        if return_pred:
            return pl.struct(pred=beta * xx + alpha, resid=yy - (beta * xx + alpha)).alias(
                "lr_pred"
            )
        else:
            return (beta.append(alpha)).implode().alias("coeffs")
    else:
        if weights is None:
            beta = xx.dot(yy) / xx.dot(xx)
        else:
            w = lr_formula(weights)
            beta = w.dot(xx * yy) / w.dot(xx.pow(2))

        if return_pred:
            return pl.struct(pred=beta * xx, resid=yy - (beta * xx)).alias("lr_pred")
        else:
            return beta.implode().alias("coeffs")


def lin_reg(
    *x: str | pl.Expr,
    target: str | pl.Expr | List[str | pl.Expr],
    add_bias: bool = False,
    weights: str | pl.Expr | None = None,
    return_pred: bool = False,
    l1_reg: float = 0.0,
    l2_reg: float = 0.0,
    tol: float = 1e-5,
    solver: LRSolverMethods = "qr",
    null_policy: NullPolicy = "skip",
    positive: bool = False,
) -> pl.Expr:
    """
    Computes linear regression solution to the equation Ax = y where y is the target (or multiple targets).
    If l1_reg is > 0, then this performs Lasso regression. If l2_reg is > 0, this performs Ridge regression.
    If both are > 0, then this is elastic net regression. If none of the cases above is true, as is the default case,
    then a normal regression will be performed.

    If add_bias is true, it will be the last coefficient in the output and output will have len(variables) + 1.

    If you only want to do simple lstsq (one predictive x variable and one target) and null policy doesn't matter,
    then `simple_lin_reg` is a faster alternative.

    Parameters
    ----------
    x
        The variables used to predict target
    target
        The target variable, or a list of targets for a multi-target linear regression
    add_bias
        Whether to add a bias term
    weights
        Whether to perform a weighted linear regression or not. If this is weighted, then it will ignore
        l1_reg or l2_reg parameters. This doesn't work if this is multi-target.
    return_pred
        If true, return prediction and residue. If false, return coefficients. Note that
        for coefficients, it reduces to one output (like max/min), but for predictions and
        residue, it will return the same number of rows as in input.
    l1_reg
        Regularization factor for Lasso. Should be nonzero when method = l1.
        This is ignored if this is multi-target.
    l2_reg
        Regularization factor for Ridge. Should be nonzero when method = l2.
    tol
        For Lasso or elastic net regression, if maximum coordinate update is < tol, the algorithm is considered
        to have converged. If not, it will run for at most 2000 iterations. This doesn't work if this is multi-target.
    solver
        Only applies when this is normal or l2 regression. One of ['svd', 'qr'].
        Both 'svd' and 'qr' can handle rank deficient cases relatively well.
    null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
        One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
        fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
        the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
        columns. If this is multi-target, fill will fail if there are nulls in any of the targets.
    positive
        If true, this will perform non-negative linear regression. Not used in multi-target case.
    """

    if isinstance(target, list):
        n_targets = len(target)
        if n_targets == 0:
            raise ValueError("If `target` is a list, it cannot be empty.")
        elif n_targets == 1:
            return lin_reg(
                *x,
                target=target[0],
                add_bias=add_bias,
                weights=weights,
                return_pred=return_pred,
                l1_reg=l1_reg,
                l2_reg=l2_reg,
                tol=tol,
                solver=solver,
                null_policy=null_policy,
            )
        else:
            cols = [lr_formula(t).alias(f"target_{i}") for i, t in enumerate(target)]
            multi_target_lr_kwargs = {
                "bias": add_bias,
                "null_policy": null_policy,
                "solver": solver,
                "last_target_idx": n_targets,
                "l2_reg": l2_reg,
            }
            cols.extend(lr_formula(z) for z in x)
            if return_pred:
                return pl_plugin(
                    symbol=cfg._which_lin_reg("pl_lstsq_multi_pred"),
                    args=cols,
                    kwargs=multi_target_lr_kwargs,
                    pass_name_to_apply=True,
                ).alias("lr_pred")
            else:
                return pl_plugin(
                    symbol=cfg._which_lin_reg("pl_lstsq_multi"),
                    args=cols,
                    kwargs=multi_target_lr_kwargs,
                    returns_scalar=True,
                    pass_name_to_apply=True,
                ).alias("coeffs")
    else:
        weighted = weights is not None
        lr_kwargs = {
            "bias": add_bias,
            "null_policy": null_policy,
            "l1_reg": l1_reg,
            "l2_reg": l2_reg,
            "solver": solver,
            "tol": tol,
            "weighted": weighted,
            "positive": positive,
        }

        if weighted:
            cols = [lr_formula(weights).cast(pl.Float64).rechunk(), lr_formula(target)]
        else:
            cols = [lr_formula(target)]

        cols.extend(lr_formula(z) for z in x)

        if return_pred:
            return pl_plugin(
                symbol=cfg._which_lin_reg("pl_lstsq_pred"),
                args=cols,
                kwargs=lr_kwargs,
                pass_name_to_apply=True,
            ).alias("lr_pred")
        else:
            return pl_plugin(
                symbol=cfg._which_lin_reg("pl_lstsq"),
                args=cols,
                kwargs=lr_kwargs,
                returns_scalar=True,
                pass_name_to_apply=True,
            ).alias("coeffs")


def lin_reg_w_rcond(
    *x: str | pl.Expr,
    target: str | pl.Expr,
    add_bias: bool = False,
    rcond: float = 0.0,
    l2_reg: float = 0.0,
    null_policy: NullPolicy = "raise",
) -> pl.Expr:
    """
    Uses SVD to compute linear regression. During the process, singular values will be set to 0
    if it is smaller than rcond * max singular value (of X). This will return the coefficients as well
    as singular values of X as the output. The number of nonzero singular values is the rank of X.

    Note: the singular values return will be the values before applying the rcond cut off.

    Parameters
    ----------
    x
        The variables used to predict target
    target
        The target variable
    add_bias
        Whether to add a bias term
    rcond
        Cut-off ratio for small singular values. If rcond < machine precision * MAX(M,N),
        it will be set to machine precision * MAX(M,N).
    l2_reg
        The L2 regularization factor. If this is > 0, then a Ridge regression will be performed.
    null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
        One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
        fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
        the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
        columns.
    """

    cols = [lr_formula(target)]
    cols.extend(lr_formula(z) for z in x)
    lr_kwargs = {
        "bias": add_bias,
        "null_policy": null_policy,
        "l1_reg": 0.0,
        "l2_reg": l2_reg,
        "solver": "",
        "tol": abs(rcond),
    }
    return pl_plugin(
        symbol=cfg._which_lin_reg("pl_lstsq_w_rcond"),
        args=cols,
        kwargs=lr_kwargs,
        pass_name_to_apply=True,
    )


def recursive_lin_reg(
    *x: str | pl.Expr,
    target: str | pl.Expr,
    start_with: int,
    add_bias: bool = False,
    l2_reg: float = 0.0,
    null_policy: NullPolicy = "raise",
) -> pl.Expr:
    """
    Using the first `start_with` rows of data as basis, start computing the least square solutions
    by updating the betas per row. A prediction for that row will also be included in the output.
    This uses the famous Sherman-Morrison-Woodbury Formula under the hood.

    Note: You have to be careful about the order of data when using this in aggregation contexts.

    Parameters
    ----------
    x:
        The variables used to predict target
    target:
        The target variable
    start_with:
        Must be >= 1. You `start_with` n rows of data to train the first linear regression. If `start_with` = N,
        the first N-1 rows will be null. If you start with N < # features, result will be numerically very
        unstable and potentially wrong.
    add_bias
        Whether to add a bias term
    l2_reg
        The L2 regularization factor. If this is > 0, then a Ridge regression will be performed.
    null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
        One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
        fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
        the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
        columns. If null_policy is `skip` or `fill`, and nulls actually exist, it will keep skipping until we have
        scanned `start_at` many valid rows. And if subsequently we get a row with null values, then null will
        be returned for that row.
    """

    if start_with < 1:
        raise ValueError("You must start with >= 1 rows for recursive lstsq.")

    cols = [lr_formula(target)]
    features = [lr_formula(z) for z in x]
    if len(features) > start_with:
        warnings.warn(
            "# features > number of rows for the initial fit. Outputs may be off.", stacklevel=2
        )

    cols.extend(features)
    kwargs = {
        "null_policy": null_policy,
        "n": start_with,
        "bias": add_bias,
        "lambda": abs(l2_reg),
        "min_size": 0,  # Not used for recursive
    }
    return pl_plugin(
        symbol=cfg._which_lin_reg("pl_recursive_lstsq"),
        args=cols,
        kwargs=kwargs,
        pass_name_to_apply=True,
    )


def rolling_lin_reg(
    *x: str | pl.Expr,
    target: str | pl.Expr,
    window_size: int,
    add_bias: bool = False,
    l2_reg: float = 0.0,
    min_valid_rows: int | None = None,
    null_policy: NullPolicy = "raise",
) -> pl.Expr:
    """
    Using every `window_size` rows of data as feature matrix, and computes least square solutions
    by rolling the window. A prediction for that row will also be included in the output.
    This uses the famous Sherman-Morrison-Woodbury Formula under the hood.

    Note: You have to be careful about the order of data when using this in aggregation contexts.
    Rows with null will not contribute to the update, so appropriate null-filling beforehand needs to be done.

    Parameters
    ----------
    x
        The variables used to predict target
    target
        The target variable
    window_size
        Must be >= 2. Window size for the rolling regression
    add_bias
        Whether to add a bias term
    l2_reg
        The L2 regularization factor. If this is > 0, then a Ridge regression will be performed.
    min_valid_rows
        Minimum number of valid rows to evaluate the model. This is only used when null policy is `skip`. E.g.
        if there are nulls in the windows, the window must have at least `min_valid_rows` valid rows in order to
        produce a result. Otherwise, null will be returned.
    null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
        One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
        fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: For
        rolling lstsq, null-fill only works when target doesn't have nulls, and WILL NOT drop rows where the
        target is null.
    """

    if window_size < 2:
        raise ValueError("`window_size` must be >= 2.")

    cols = [lr_formula(target)]
    features = [lr_formula(z) for z in x]
    if len(features) > window_size:
        raise ValueError("# features > window size. Linear regression is not well-defined.")

    if min_valid_rows is None:
        min_size = min(len(features), window_size)
    else:
        if min_valid_rows < len(features):
            warnings.warn(
                "# features > min_window_size. Linear regression may not always be well-defined.",
                stacklevel=2,
            )
        min_size = min_valid_rows

    cols.extend(features)
    kwargs = {
        "null_policy": null_policy,
        "n": window_size,
        "bias": add_bias,
        "lambda": abs(l2_reg),
        "min_size": min_size,
    }
    return pl_plugin(
        symbol=cfg._which_lin_reg("pl_rolling_lstsq"),
        args=cols,
        kwargs=kwargs,
        pass_name_to_apply=True,
    )


def lin_reg_report(
    *x: str | pl.Expr,
    target: str | pl.Expr,
    weights: str | pl.Expr | None = None,
    add_bias: bool = False,
    null_policy: NullPolicy = "raise",
    std_err: Literal["se", "hc0", "hc1", "hc2", "hc3"] = "se",
) -> pl.Expr:
    """
    Creates an ordinary least square report with more stats about each coefficient.

    Note: if columns are not linearly independent, some numerical issue may occur. This uses
    the closed form solution to compute the least square report.

    Parameters
    ----------
    x
        The variables used to predict target
    target
        The target variable
    weights
        If not None, this will then compute the stats for a weights least square.
    add_bias
        Whether to add a bias term. If bias is added, it is always the last feature.
    null_policy: Literal['raise', 'skip', 'zero', 'one', 'ignore']
        One of options shown here, but you can also pass in any numeric string. E.g you may pass '1.25' to mean
        fill nulls with 1.25. If the string cannot be converted to a float, an error will be thrown. Note: if
        the target column has null, the rows with nulls will always be dropped. Null-fill only applies to non-target
        columns.
    std_err
        One of "se", "hc0", "hc1", "hc2", "hc3", where "se" means we compute the standard error
        under the assumption of homoskedasticity, and the hc options are different options for
        heteroskedasticity. The hc0-hc3 are called Heteroskedasticity-Consistent Standard Errors, and their
        formulas can be found here: https://jslsoc.sitehost.iu.edu/files_research/testing_tests/hccm/00TAS.pdf.
        This won't be used if weights are used (The author is not super familiar with the theory). If any other
        string is provided, it will default to "se".
    """

    lr_kwargs = {
        "bias": add_bias,
        "null_policy": null_policy,
        "l1_reg": 0.0,
        "l2_reg": 0.0,
        "solver": "qr",
        "tol": 0.0,
        "std_err": std_err.lower(),
    }

    t = lr_formula(target)
    if weights is None:
        cols = [t.var(), t]
        cols.extend(lr_formula(z) for z in x)
        symbol = cfg._which_lin_reg("pl_lin_reg_report")

    else:
        w = lr_formula(weights)
        cols = [w.cast(pl.Float64).rechunk(), t.var(), t]
        cols.extend(lr_formula(z) for z in x)
        symbol = cfg._which_lin_reg("pl_wls_report")

    return pl_plugin(
        symbol=symbol,
        args=cols,
        kwargs=lr_kwargs,
        changes_length=True,
        pass_name_to_apply=True,
    )



LIN_REG_EXPR_F64 = True
"""
If true, all linear regression expression will use f64 as the default data type
in the underlying implementation. If fase, f32 will be used. This only controls
linear regression expressions.

The memory footprint will be smaller, but it is possible to have slower speed than f64 for 
multiple reasons:
1. If input data is already in f64, then using f32 will incur additional casts, slowing
down the process.
2. If input data is not big enough, there won't be any noticeable difference in runtime.
"""


def _which_lin_reg(x: str) -> str:
    return x if LIN_REG_EXPR_F64 else f"{x}_f32"



import polars_ds as pds
import numpy as np
from polars_ds.spatial import KDTree as KDT
from scipy.spatial import KDTree


def test_kdtree():
    size = 2000
    df = pds.frame(size=size).with_columns(*(pds.random().alias(f"var{i}") for i in range(3)))
    X = df.select(f"var{i}" for i in range(3)).to_numpy(order="c")

    pds_tree = KDT(X, distance="l2")
    scipy_tree = KDTree(X, copy_data=True)

    distances_pds, indices_pds = pds_tree.knn(X, k=10, parallel=False)
    distances_scipy, indices_scipy = scipy_tree.query(X, k=10, p=2)

    assert np.all(distances_pds == distances_scipy)
    assert np.all(indices_pds.astype(np.int64) == indices_scipy.astype(np.int64))

    within_pds = pds_tree.within(X, r=0.1, sort=False)
    within_scipy = scipy_tree.query_ball_point(X, r=0.1, p=2, return_sorted=False)
    assert all(set(x) == set(y) for x, y in zip(within_pds, within_scipy))

    within_count = pds_tree.within_count(X, r=0.1)
    assert all(int(n) == len(pts) for n, pts in zip(within_count, within_pds))



import polars as pl
import numpy as np
from typing import Any, Callable
import polars_ds as pds

# Everything in __init__.py of polars_ds that this shouldn't be able to call
CANNOT_CALL = {
    "frame",
    "str_to_expr",
    "pl",
    "annotations",
    "__version__",
    "warn_len_compare",
    "E",
    "Config",
}

__all__ = ["compat"]


class _Compat:
    def __init__(self):
        self.return_numpy = False

    @staticmethod
    def _try_into_series(x: Any, name: str) -> Any:
        """
        Try to map the input to a Polars Series by going through a NumPy array. If
        this is not possible, return the original input.
        """
        if isinstance(x, np.ndarray):
            return pl.lit(pl.Series(name=name, values=x))
        elif isinstance(x, pl.Series):
            return pl.lit(x)
        elif hasattr(x, "__array__"):
            return pl.lit(pl.Series(name=name, values=x.__array__()))
        else:
            return x

    def __getattr__(self, name: str) -> Any:
        if name in CANNOT_CALL:
            raise ValueError(f"`{name}` exists but doesn't work in compat mode.")

        func = getattr(pds, name)

        def compat_wrapper(*args, **kwargs) -> Callable:
            positionals = list(args)
            if len(positionals) <= 0:
                raise ValueError("There must be at least 1 positional argument!")

            first_expr = positionals[0]
            new_args = (_Compat._try_into_series(x, name=str(i)) for i, x in enumerate(positionals))
            new_kwargs = {n: _Compat._try_into_series(v, name=n) for n, v in kwargs.items()}

            output = pl.select(func(*new_args, **new_kwargs).alias("__output__")).drop_in_place(
                "__output__"
            )

            if self.return_numpy or isinstance(first_expr, np.ndarray):
                return output.to_numpy()
            else:
                return first_expr.__class__(output.to_list()).rename("")

        return compat_wrapper


compat: _Compat = _Compat()



import polars as pl
import polars_ds as pds
import pytest
import numpy as np
from sklearn.linear_model import LinearRegression

from polars_ds.linear_models import OnlineLR, ElasticNet, LR


def test_lr_null_policies_for_np():
    from polars_ds.linear_models import _handle_nans_in_np

    size = 5_000
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_row_index()
        .with_columns(
            x1=pl.when(pl.col("x1") > 0.5).then(None).otherwise(pl.col("x1")),
            y=pl.col("x1") + pl.col("x2") * 0.2 - 0.3 * pl.col("x3"),
        )
        .with_columns(is_null=pl.col("x1").is_null())
    )
    nulls = df.select("is_null").to_numpy().flatten()
    x = df.select("x1", "x2", "x3").to_numpy()
    y = df.select("y").to_numpy()

    x_nan, _ = _handle_nans_in_np(x, y, "ignore")
    assert np.all(np.isnan(x_nan[nulls][:, 0]))

    with pytest.raises(Exception) as exc_info:
        _handle_nans_in_np(x, y, "raise")
        assert str(exc_info.value) == "Nulls found in X or y."

    x_skipped, _ = _handle_nans_in_np(x, y, "skip")
    assert np.all(x_skipped == x[~nulls])

    x_zeroed, _ = _handle_nans_in_np(x, y, "zero")
    assert np.all(
        x_zeroed[nulls][:, 0] == 0.0
    )  # checking out the first column because only that has nulls

    x_one, _ = _handle_nans_in_np(x, y, "one")
    assert np.all(
        x_one[nulls][:, 0] == 1.0
    )  # checking out the first column because only that has nulls


@pytest.mark.parametrize("solver", ["svd", "cholesky", "qr"])
def test_lr(solver):
    ols = LR(False, 0.0, solver)
    df = (
        pds.frame(size=5000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_row_index()
        .with_columns(
            y=pl.col("x1") + pl.col("x2") * 0.2 - 0.3 * pl.col("x3") + pds.random() * 0.0001
        )
    )
    X = df.select("x1", "x2", "x3").to_numpy()
    y = df.select("y").to_numpy()

    ols.fit(X, y)
    coeffs = ols.coeffs()
    sk_ols = LinearRegression(fit_intercept=False)
    sk_ols.fit(X, y)
    sklearn_coeffs = sk_ols.coef_
    assert np.all(np.abs(coeffs - sklearn_coeffs) < 1e-6)


def test_online_lr():
    size = 5000
    df = (
        pds.frame(size=size)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_row_index()
        .with_columns(
            y=pl.col("x1") + pl.col("x2") * 0.2 - 0.3 * pl.col("x3") + pds.random() * 0.0001
        )
    )
    X = df.select("x1", "x2", "x3").to_numpy()
    y = df.select("y").to_numpy()

    olr = OnlineLR()  # no bias, normal
    sk_lr = LinearRegression(fit_intercept=False)

    olr.fit(X[:10], y[:10])
    coeffs = olr.coeffs()
    sk_lr.fit(X[:10], y[:10])
    sklearn_coeffs = sk_lr.coef_

    pred = olr.predict(X[:10]).flatten()
    sk_pred = sk_lr.predict(X[:10]).flatten()
    assert np.all(np.abs(pred - sk_pred) < 1e-6)
    assert np.all(np.abs(coeffs - sklearn_coeffs) < 1e-6)

    for i in range(10, 20):
        olr.update(X[i], y[i])
        coeffs = olr.coeffs()
        sk_lr = LinearRegression(fit_intercept=False)
        sk_lr.fit(X[: i + 1], y[: i + 1])
        sklearn_coeffs = sk_lr.coef_
        assert np.all(np.abs(coeffs - sklearn_coeffs) < 1e-6)


def _test_elastic_net(add_bias: bool = False):
    import sklearn.linear_model as lm

    l1_reg = 0.1
    l2_reg = 0.1
    alpha = l1_reg + l2_reg
    l1_ratio = l1_reg / (l1_reg + l2_reg)

    df = (
        pds.frame(size=5000)
        .select(
            pds.random(0.0, 1.0).alias("x1"),
            pds.random(0.0, 1.0).alias("x2"),
            pds.random(0.0, 1.0).alias("x3"),
        )
        .with_row_index()
        .with_columns(
            y=pl.col("x1") + pl.col("x2") * 0.2 - 0.3 * pl.col("x3"),
        )
        .with_columns(is_null=pl.col("x1").is_null())
    )

    X = df.select("x1", "x2", "x3").to_numpy()
    y = df.select("y").to_numpy()
    en = ElasticNet(l1_reg=l1_reg, l2_reg=l2_reg, has_bias=add_bias)
    elastic = lm.ElasticNet(alpha=alpha, l1_ratio=l1_ratio, fit_intercept=add_bias)

    en.fit(X, y)
    pds_res = en.coeffs()
    elastic.fit(X, y)
    sklearn_res = elastic.coef_
    assert np.all(np.abs(pds_res - sklearn_res) < 1e-4)

    if add_bias is True:
        pds_bias = en.bias()
        sklearn_bias = elastic.intercept_
        assert np.all(np.abs(pds_bias - sklearn_bias) < 1e-4)


def test_elastic_net():
    _test_elastic_net(add_bias=False)
    _test_elastic_net(add_bias=True)


def test_glm():
    import statsmodels.api as sm
    from polars_ds.linear_models import GLM

    data = sm.datasets.scotland.load()
    # add_bias = True

    data.exog = sm.add_constant(data.exog)
    df = data.exog
    df["y"] = data.endog
    # X1 contains constant term, X2 doesn't
    X1 = df[[c for c in df.columns if c != "y"]].to_numpy()
    X2 = df[[c for c in df.columns if c not in ("y", "const")]].to_numpy()
    y = df["y"].to_numpy()

    glm = sm.GLM(y, X1, family=sm.families.Gamma())
    glm_results = glm.fit()
    params = glm_results.params
    sm_bias = params[0]
    sm_coeffs = params[1:]

    pds_glm = GLM(solver="irls", add_bias=True, family="gamma")

    pds_glm.fit(X2, y.reshape((-1, 1)))

    pds_bias = pds_glm.bias()
    pds_coeffs = pds_glm.coeffs()

    assert abs(sm_bias - pds_bias) < 1e-6
    np.testing.assert_allclose(sm_coeffs, pds_coeffs, rtol=1e-5)



import random
import string
from pathlib import Path

import polars as pl
from faker import Faker

random.seed(208)

BASE_PATH = Path(__file__).resolve().parents[0]

INT32_RANGE = (-(2**31), 2**31 - 1)


UNICODE_ALPHABET = [
    chr(code_point)
    for current_range in [
        (0x0021, 0x0021),
        (0x0023, 0x0026),
        (0x0028, 0x007E),
        (0x00A1, 0x00AC),
        (0x00AE, 0x00FF),
        (0x0100, 0x017F),
        (0x0180, 0x024F),
        (0x2C60, 0x2C7F),
        (0x16A0, 0x16F0),
        (0x0370, 0x0377),
        (0x037A, 0x037E),
        (0x0384, 0x038A),
        (0x038C, 0x038C),
    ]
    for code_point in range(current_range[0], current_range[1] + 1)
]


def random_unicode(length: int) -> str:
    # https://stackoverflow.com/questions/1477294/generate-random-utf-8-string-in-python
    return "".join(random.choices(UNICODE_ALPHABET, k=length))


def random_ascii(length: int) -> str:
    return "".join(random.choices(string.printable, k=length))


class DataGenerator:
    def __init__(self, height: int):
        self.height = height

    def random_unicode_column(self, min_length: int, max_length: int) -> list[str]:
        column = []
        for _ in range(self.height):
            length = random.randint(min_length, max_length)
            column.append(random_unicode(length))

        return column

    def random_ascii_column(self, min_length: int, max_length: int) -> list[str]:
        column = []
        for _ in range(self.height):
            length = random.randint(min_length, max_length)
            column.append(random_ascii(length))

        return column

    def random_address_column(self) -> list[str]:
        column = []
        fake = Faker()
        for i in range(self.height):
            Faker.seed(i)
            column.append(fake.address())

        return column

    def random_integer_column(self, min_value: int, max_value: int) -> list[int]:
        return [random.randint(min_value, max_value) for _ in range(self.height)]

    def random_float_column(self, min_value: float, max_value: float) -> list[float]:
        return [random.uniform(min_value, max_value) for _ in range(self.height)]

    def generate(self) -> pl.DataFrame:
        min_length = 1
        max_length = 50

        return pl.DataFrame(
            {
                "RANDOM_STRING": self.random_unicode_column(min_length, max_length),
                "RANDOM_ASCII": self.random_ascii_column(min_length, max_length),
                "RANDOM_ADDRESS": self.random_address_column(),
            }
        )


def main():
    DataGenerator(1_000_000).generate().write_parquet(BASE_PATH / "benchmark_df.parquet")


if __name__ == "__main__":
    main()
